"""Unit tests for routing session makers.

Tests the factory classes that create routing sessions.
"""

from typing import Any
from unittest.mock import MagicMock

import pytest
from sqlalchemy import Engine
from sqlalchemy.ext.asyncio import AsyncEngine

from advanced_alchemy.config.routing import RoutingConfig, RoutingStrategy
from advanced_alchemy.routing.maker import (
    RoutingAsyncSessionMaker,
    RoutingSyncSessionMaker,
    dispose_session_maker_async,
    dispose_session_maker_sync,
)
from advanced_alchemy.routing.selectors import RandomSelector, RoundRobinSelector
from advanced_alchemy.routing.session import RoutingSyncSession


@pytest.fixture
def routing_config() -> RoutingConfig:
    """Create a routing config with two replicas."""
    return RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=[
            "postgresql://replica1:5432/db",
            "postgresql://replica2:5432/db",
        ],
    )


def test_sync_session_maker_initialization(routing_config: RoutingConfig) -> None:
    """Test RoutingSyncSessionMaker initialization."""

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        engine = MagicMock(spec=["dispose"])
        engine.url = url
        return engine

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 2
    assert maker.primary_engine.url == "postgresql://primary:5432/db"
    assert maker.replica_engines[0].url == "postgresql://replica1:5432/db"
    assert maker.replica_engines[1].url == "postgresql://replica2:5432/db"


def test_sync_session_maker_creates_round_robin_selector_by_default(
    routing_config: RoutingConfig,
) -> None:
    """Test that RoutingSyncSessionMaker creates RoundRobinSelector by default."""

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        engine = MagicMock(spec=["dispose"])
        engine.url = url
        return engine

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_engine,
    )

    session = maker()
    assert isinstance(session._selectors[routing_config.read_group], RoundRobinSelector)


def test_sync_session_maker_creates_random_selector(routing_config: RoutingConfig) -> None:
    """Test that RoutingSyncSessionMaker creates RandomSelector when configured."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=["postgresql://replica1:5432/db"],
        routing_strategy=RoutingStrategy.RANDOM,
    )

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        engine = MagicMock(spec=["dispose"])
        engine.url = url
        return engine

    maker = RoutingSyncSessionMaker(
        routing_config=config,
        create_engine_callable=create_mock_engine,
    )

    session = maker()
    assert isinstance(session._selectors[config.read_group], RandomSelector)


def test_sync_session_maker_call_creates_session(routing_config: RoutingConfig) -> None:
    """Test that calling the maker creates a RoutingSyncSession."""

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        return MagicMock(spec=["dispose"])

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_engine,
    )

    session = maker()

    assert isinstance(session, RoutingSyncSession)
    assert session._routing_config is routing_config


def test_sync_session_maker_passes_engine_config() -> None:
    """Test that RoutingSyncSessionMaker passes engine config to create_engine."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=["postgresql://replica1:5432/db"],
    )

    engine_config = {"pool_size": 10, "max_overflow": 20}

    mock_create_engine = MagicMock(return_value=MagicMock(spec=["dispose"]))

    RoutingSyncSessionMaker(
        routing_config=config,
        engine_config=engine_config,
        create_engine_callable=mock_create_engine,
    )

    assert mock_create_engine.call_count == 2
    for call in mock_create_engine.call_args_list:
        _, kwargs = call
        assert kwargs.get("pool_size") == 10
        assert kwargs.get("max_overflow") == 20


def test_sync_session_maker_passes_session_config(routing_config: RoutingConfig) -> None:
    """Test that RoutingSyncSessionMaker passes session config to session."""

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        return MagicMock(spec=["dispose"])

    session_config = {"expire_on_commit": False, "autoflush": False}

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        engine_config={},
        session_config=session_config,
        create_engine_callable=create_mock_engine,
    )

    session = maker()

    assert session.expire_on_commit is False
    assert session.autoflush is False


def test_sync_session_maker_close_all_disposes_engines(routing_config: RoutingConfig) -> None:
    """Test that close_all disposes all engines."""
    mock_engines = []

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        engine = MagicMock(spec=["dispose"])
        mock_engines.append(engine)
        return engine

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_engine,
    )

    maker.close_all()

    assert len(mock_engines) == 3
    for engine in mock_engines:
        engine.dispose.assert_called_once()


def test_sync_session_maker_handles_engine_creation_errors() -> None:
    """Test that RoutingSyncSessionMaker handles TypeError from unsupported engine options."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=["postgresql://replica1:5432/db"],
    )

    call_count = 0

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        nonlocal call_count
        call_count += 1

        if "json_serializer" in kwargs:
            raise TypeError("json_serializer not supported")

        engine = MagicMock(spec=["dispose"])
        engine.url = url
        return engine

    engine_config = {"pool_size": 10, "json_serializer": lambda x: x}

    maker = RoutingSyncSessionMaker(
        routing_config=config,
        engine_config=engine_config,
        create_engine_callable=create_mock_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 1


def test_async_session_maker_initialization(routing_config: RoutingConfig) -> None:
    """Test RoutingAsyncSessionMaker initialization."""

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.url = url
        engine.sync_engine = MagicMock()
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_async_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 2
    assert maker.primary_engine.url == "postgresql://primary:5432/db"


def test_async_session_maker_creates_round_robin_selector_by_default(
    routing_config: RoutingConfig,
) -> None:
    """Test that RoutingAsyncSessionMaker creates RoundRobinSelector by default."""

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_async_engine,
    )

    assert isinstance(maker._selectors[routing_config.read_group], RoundRobinSelector)


def test_async_session_maker_creates_random_selector(routing_config: RoutingConfig) -> None:
    """Test that RoutingAsyncSessionMaker creates RandomSelector when configured."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=["postgresql://replica1:5432/db"],
        routing_strategy=RoutingStrategy.RANDOM,
    )

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=config,
        create_engine_callable=create_mock_async_engine,
    )

    assert isinstance(maker._selectors[config.read_group], RandomSelector)


def test_async_session_maker_call_creates_session(routing_config: RoutingConfig) -> None:
    """Test that calling the async maker would create a RoutingAsyncSession.

    Note: We can't fully test this with mocks because RoutingAsyncSession
    initialization requires special engine handling. This test is covered
    in integration tests instead.
    """

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_async_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 2


def test_async_session_maker_passes_engine_config() -> None:
    """Test that RoutingAsyncSessionMaker passes engine config to create_async_engine."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=["postgresql://replica1:5432/db"],
    )

    engine_config = {"pool_size": 10, "max_overflow": 20}

    mock_create_engine = MagicMock()
    mock_create_engine.return_value = MagicMock(spec=["dispose", "sync_engine"])
    mock_create_engine.return_value.sync_engine = MagicMock()

    RoutingAsyncSessionMaker(
        routing_config=config,
        engine_config=engine_config,
        create_engine_callable=mock_create_engine,
    )

    assert mock_create_engine.call_count == 2
    for call in mock_create_engine.call_args_list:
        _, kwargs = call
        assert kwargs.get("pool_size") == 10
        assert kwargs.get("max_overflow") == 20


@pytest.mark.asyncio
async def test_async_session_maker_close_all_disposes_engines(
    routing_config: RoutingConfig,
) -> None:
    """Test that close_all disposes all async engines."""
    import asyncio

    mock_engines = []

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()

        async def mock_dispose() -> None:
            pass

        engine.dispose = MagicMock(side_effect=lambda: asyncio.create_task(mock_dispose()))
        mock_engines.append(engine)
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_async_engine,
    )

    await maker.close_all()

    assert len(mock_engines) == 3
    for engine in mock_engines:
        engine.dispose.assert_called_once()


def test_async_session_maker_handles_engine_creation_errors() -> None:
    """Test that RoutingAsyncSessionMaker handles TypeError from unsupported options."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=["postgresql://replica1:5432/db"],
    )

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        if "json_serializer" in kwargs:
            raise TypeError("json_serializer not supported")

        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()
        engine.url = url
        return engine

    engine_config = {"pool_size": 10, "json_serializer": lambda x: x}

    maker = RoutingAsyncSessionMaker(
        routing_config=config,
        engine_config=engine_config,
        create_engine_callable=create_mock_async_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 1


def test_sync_session_maker_no_replicas(routing_config: RoutingConfig) -> None:
    """Test RoutingSyncSessionMaker with no replicas."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=[],
    )

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        return MagicMock(spec=["dispose"])

    maker = RoutingSyncSessionMaker(
        routing_config=config,
        create_engine_callable=create_mock_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 0


def test_async_session_maker_no_replicas(routing_config: RoutingConfig) -> None:
    """Test RoutingAsyncSessionMaker with no replicas."""
    config = RoutingConfig(
        primary_connection_string="postgresql://primary:5432/db",
        read_replicas=[],
    )

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=config,
        create_engine_callable=create_mock_async_engine,
    )

    assert maker.primary_engine is not None
    assert len(maker.replica_engines) == 0


def test_sync_session_maker_removes_bind_from_session_config(
    routing_config: RoutingConfig,
) -> None:
    """Test that session maker removes 'bind' from session config."""

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        return MagicMock(spec=["dispose"])

    session_config = {"bind": MagicMock(), "expire_on_commit": False}

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        session_config=session_config,
        create_engine_callable=create_mock_engine,
    )

    session = maker()
    assert isinstance(session, RoutingSyncSession)


def test_async_session_maker_stores_session_config(
    routing_config: RoutingConfig,
) -> None:
    """Test that async session maker stores session config correctly."""

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()
        return engine

    session_config = {"expire_on_commit": False}

    maker = RoutingAsyncSessionMaker(
        routing_config=routing_config,
        session_config=session_config,
        create_engine_callable=create_mock_async_engine,
    )

    assert maker._session_config == session_config


def test_dispose_session_maker_sync_disposes_routing_sync_maker(
    routing_config: RoutingConfig,
) -> None:
    """dispose_session_maker_sync calls close_all on a sync routing maker."""
    mock_engines: list[Any] = []

    def create_mock_engine(url: str, **kwargs: Any) -> Engine:
        engine = MagicMock(spec=["dispose"])
        mock_engines.append(engine)
        return engine

    maker = RoutingSyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_engine,
    )

    dispose_session_maker_sync(maker)

    assert len(mock_engines) == 3
    for engine in mock_engines:
        engine.dispose.assert_called_once()


def test_dispose_session_maker_sync_noop_for_non_routing_maker() -> None:
    """dispose_session_maker_sync is a no-op for non-routing makers."""
    plain = MagicMock()
    dispose_session_maker_sync(plain)
    plain.close_all.assert_not_called()


def test_dispose_session_maker_sync_noop_for_none() -> None:
    """dispose_session_maker_sync handles a None session maker without raising."""
    dispose_session_maker_sync(None)


@pytest.mark.asyncio
async def test_dispose_session_maker_async_disposes_routing_async_maker(
    routing_config: RoutingConfig,
) -> None:
    """dispose_session_maker_async calls close_all on an async routing maker."""
    mock_engines: list[Any] = []

    def create_mock_async_engine(url: str, **kwargs: Any) -> AsyncEngine:
        engine = MagicMock(spec=["dispose", "sync_engine"])
        engine.sync_engine = MagicMock()

        async def _dispose() -> None:
            return None

        engine.dispose.side_effect = _dispose
        mock_engines.append(engine)
        return engine

    maker = RoutingAsyncSessionMaker(
        routing_config=routing_config,
        create_engine_callable=create_mock_async_engine,
    )

    await dispose_session_maker_async(maker)

    assert len(mock_engines) == 3
    for engine in mock_engines:
        engine.dispose.assert_called_once()


@pytest.mark.asyncio
async def test_dispose_session_maker_async_noop_for_non_routing_maker() -> None:
    """dispose_session_maker_async is a no-op for non-routing makers."""
    plain = MagicMock()
    await dispose_session_maker_async(plain)
    plain.close_all.assert_not_called()


@pytest.mark.asyncio
async def test_dispose_session_maker_async_noop_for_none() -> None:
    """dispose_session_maker_async handles a None session maker without raising."""
    await dispose_session_maker_async(None)



"""Tests for attrs integration in Advanced Alchemy services."""

from __future__ import annotations

from typing import Any, Optional

import pytest
from attrs import define, field

from advanced_alchemy.utils.serialization import (
    ATTRS_INSTALLED,
    AttrsInstance,
    is_attrs_instance,
    is_attrs_instance_with_field,
    is_attrs_instance_without_field,
    is_attrs_schema,
    is_schema,
    is_schema_with_field,
    is_schema_without_field,
    schema_dump,
)

pytestmark = [
    pytest.mark.unit,
]

# attrs test classes and fixtures


@define
class SimpleAttrsInstance:
    """Simple attrs class for testing."""

    name: str
    age: int


@define
class AttrsWithOptional:
    """attrs class with optional fields."""

    name: str
    email: Optional[str] = None
    active: bool = True


@define
class AttrsWithDefaults:
    """attrs class with field defaults."""

    title: str
    count: int = field(default=0)
    tags: list[str] = field(factory=list)


@define
class NestedAttrsInstance:
    """attrs class with nested attrs."""

    user: SimpleAttrsInstance
    metadata: dict[str, Any] = field(factory=dict)


class TestAttrsDetection:
    """Test attrs class detection functions."""

    def test_attrs_installed_flag(self) -> None:
        """Test that ATTRS_INSTALLED flag is a boolean."""
        assert isinstance(ATTRS_INSTALLED, bool)

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_attrs_instance_with_attrs_instance(self) -> None:
        """Test is_attrs_instance with actual attrs class instance."""
        instance = SimpleAttrsInstance(name="test", age=30)
        assert is_attrs_instance(instance)

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_attrs_instance_with_complex_attrs(self) -> None:
        """Test is_attrs_instance with complex attrs instances."""
        simple_instance = SimpleAttrsInstance(name="John", age=25)
        nested_instance = NestedAttrsInstance(user=simple_instance, metadata={"role": "admin"})

        assert is_attrs_instance(simple_instance)
        assert is_attrs_instance(nested_instance)

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_attrs_schema_with_attrs_classes(self) -> None:
        """Test is_attrs_schema with attrs class types."""
        assert is_attrs_schema(SimpleAttrsInstance)
        assert is_attrs_schema(AttrsWithOptional)
        assert is_attrs_schema(AttrsWithDefaults)
        assert is_attrs_schema(NestedAttrsInstance)

    def test_is_attrs_schema_with_non_attrs_classes(self) -> None:
        """Test is_attrs_schema with non-attrs class types."""
        assert not is_attrs_schema(dict)
        assert not is_attrs_schema(list)
        assert not is_attrs_schema(str)
        assert not is_attrs_schema(int)

        class RegularClass:
            pass

        assert not is_attrs_schema(RegularClass)

    def test_is_attrs_class_with_non_attrs_instance(self) -> None:
        """Test is_attrs_class with non-attrs objects."""
        assert not is_attrs_instance({})
        assert not is_attrs_instance([])
        assert not is_attrs_instance("string")
        assert not is_attrs_instance(42)

    def test_is_attrs_class_with_regular_class(self) -> None:
        """Test is_attrs_class with regular Python class."""

        class RegularClass:
            def __init__(self, name: str) -> None:
                self.name = name

        instance = RegularClass("test")
        assert not is_attrs_instance(instance)

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_attrs_instance_with_field(self) -> None:
        """Test is_attrs_class_with_field function."""
        instance = SimpleAttrsInstance(name="test", age=30)

        assert is_attrs_instance_with_field(instance, "name")
        assert is_attrs_instance_with_field(instance, "age")
        assert not is_attrs_instance_with_field(instance, "nonexistent")

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_attrs_instance_without_field(self) -> None:
        """Test is_attrs_class_without_field function."""
        instance = SimpleAttrsInstance(name="test", age=30)

        assert not is_attrs_instance_without_field(instance, "name")
        assert not is_attrs_instance_without_field(instance, "age")
        assert is_attrs_instance_without_field(instance, "nonexistent")

    def test_is_attrs_class_field_checks_with_non_attrs(self) -> None:
        """Test field checking functions with non-attrs objects."""
        regular_obj = {"name": "test", "age": 30}

        assert not is_attrs_instance_with_field(regular_obj, "name")
        assert not is_attrs_instance_without_field(regular_obj, "name")


class TestSchemaIntegration:
    """Test attrs integration with schema functions."""

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_schema_with_attrs_class(self) -> None:
        """Test is_schema function includes attrs classes."""
        instance = SimpleAttrsInstance(name="test", age=30)
        assert is_schema(instance)

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_schema_with_field_attrs(self) -> None:
        """Test is_schema_with_field with attrs classes."""
        instance = SimpleAttrsInstance(name="test", age=30)

        assert is_schema_with_field(instance, "name")
        assert is_schema_with_field(instance, "age")
        assert not is_schema_with_field(instance, "nonexistent")

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_is_schema_without_field_attrs(self) -> None:
        """Test is_schema_without_field with attrs classes."""
        instance = SimpleAttrsInstance(name="test", age=30)

        assert not is_schema_without_field(instance, "name")
        assert not is_schema_without_field(instance, "age")
        assert is_schema_without_field(instance, "nonexistent")


class TestSchemaDump:
    """Test schema_dump function with attrs classes."""

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_schema_dump_simple_attrs(self) -> None:
        """Test schema_dump with simple attrs class."""
        instance = SimpleAttrsInstance(name="John", age=30)
        result = schema_dump(instance)

        expected = {"name": "John", "age": 30}
        assert result == expected
        assert isinstance(result, dict)

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_schema_dump_attrs_with_optional(self) -> None:
        """Test schema_dump with attrs class having optional fields."""
        instance = AttrsWithOptional(name="Jane", email="jane@example.com", active=True)
        result = schema_dump(instance)

        expected = {"name": "Jane", "email": "jane@example.com", "active": True}
        assert result == expected

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_schema_dump_attrs_with_defaults(self) -> None:
        """Test schema_dump with attrs class having field defaults."""
        instance = AttrsWithDefaults(title="Test", count=5, tags=["tag1", "tag2"])
        result = schema_dump(instance)

        expected = {"title": "Test", "count": 5, "tags": ["tag1", "tag2"]}
        assert result == expected

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_schema_dump_nested_attrs(self) -> None:
        """Test schema_dump with nested attrs classes."""
        user = SimpleAttrsInstance(name="John", age=25)
        instance = NestedAttrsInstance(user=user, metadata={"role": "admin"})
        result = schema_dump(instance)

        expected = {"user": {"name": "John", "age": 25}, "metadata": {"role": "admin"}}
        assert result == expected

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_schema_dump_attrs_collection_types(self) -> None:
        """Test schema_dump retains collection types."""
        instance = AttrsWithDefaults(title="Test", tags=["a", "b", "c"])
        result = schema_dump(instance)

        assert isinstance(result["tags"], list)
        assert result["tags"] == ["a", "b", "c"]

    def test_schema_dump_non_attrs_unchanged(self) -> None:
        """Test schema_dump with non-attrs objects remains unchanged."""
        # Dict should pass through unchanged
        dict_data = {"name": "test", "age": 30}
        assert schema_dump(dict_data) == dict_data

    def test_schema_dump_none_value(self) -> None:
        """Test that schema_dump correctly handles None values.

        schema_dump should gracefully handle None input values without raising
        AttributeError, returning None as-is through the fallback mechanism.
        This is important for nullable database fields that may contain None.
        """
        # schema_dump should handle None without raising AttributeError
        result = schema_dump(None)
        assert result is None

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_schema_dump_exclude_unset_parameter(self) -> None:
        """Test schema_dump exclude_unset parameter (attrs always includes all fields)."""
        instance = SimpleAttrsInstance(name="test", age=30)

        # attrs.asdict doesn't have exclude_unset concept, should return all fields
        result_with_unset = schema_dump(instance, exclude_unset=True)
        result_without_unset = schema_dump(instance, exclude_unset=False)

        expected = {"name": "test", "age": 30}
        assert result_with_unset == expected
        assert result_without_unset == expected


class TestAttrsInstanceProtocol:
    """Test AttrsInstance protocol."""

    def test_attrs_class_protocol_exists(self) -> None:
        """Test that AttrsInstance protocol is available."""
        assert AttrsInstance is not None

    @pytest.mark.skipif(not ATTRS_INSTALLED, reason="attrs not installed")
    def test_attrs_class_protocol_with_real_attrs(self) -> None:
        """Test AttrsInstance protocol with real attrs class."""
        instance = SimpleAttrsInstance(name="test", age=30)

        # Should have __attrs_attrs__ attribute when attrs is installed
        assert hasattr(instance.__class__, "__attrs_attrs__")

    def test_attrs_class_protocol_type_annotation(self) -> None:
        """Test AttrsInstance can be used in type annotations."""

        # This test ensures the protocol is properly defined for type checking
        def process_attrs(obj: AttrsInstance) -> dict[str, Any]:
            return {"processed": True}

        # Should not raise type errors
        assert callable(process_attrs)


class TestErrorHandling:
    """Test error handling in attrs integration."""

    @pytest.mark.skipif(ATTRS_INSTALLED, reason="attrs is installed")
    def test_attrs_functions_when_not_installed(self) -> None:
        """Test attrs functions behave correctly when attrs not installed."""
        # When attrs not installed, detection should return False
        dummy_obj = SimpleAttrsInstance(name="test", age=30)

        assert not is_attrs_instance(dummy_obj)
        assert not is_attrs_instance_with_field(dummy_obj, "name")
        assert not is_attrs_instance_without_field(dummy_obj, "name")

    @pytest.mark.skipif(ATTRS_INSTALLED, reason="attrs is installed")
    def test_schema_dump_fallback_when_attrs_not_installed(self) -> None:
        """Test schema_dump falls back to __dict__ when attrs not installed."""
        dummy_obj = SimpleAttrsInstance(name="test", age=30)
        result = schema_dump(dummy_obj)

        # Should fall back to __dict__ access
        expected = {"name": "test", "age": 30}
        assert result == expected



import asyncio
from typing import TYPE_CHECKING, cast

from alembic.autogenerate import rewriter
from sqlalchemy import pool
from sqlalchemy.ext.asyncio import AsyncEngine, async_engine_from_config

from advanced_alchemy.base import metadata_registry
from alembic import context

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection

    from advanced_alchemy.alembic.commands import AlembicCommandConfig

__all__ = ("do_run_migrations", "run_migrations_offline", "run_migrations_online")


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config: "AlembicCommandConfig" = context.config  # type: ignore
writer = rewriter.Rewriter()


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.
    """
    context.configure(
        url=config.db_url,
        target_metadata=metadata_registry.get(config.bind_key),
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        compare_type=config.compare_type,
        version_table=config.version_table_name,
        version_table_pk=config.version_table_pk,
        user_module_prefix=config.user_module_prefix,
        render_as_batch=config.render_as_batch,
        process_revision_directives=writer,
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: "Connection") -> None:
    """Run migrations."""
    context.configure(
        connection=connection,
        target_metadata=metadata_registry.get(config.bind_key),
        compare_type=config.compare_type,
        version_table=config.version_table_name,
        version_table_pk=config.version_table_pk,
        user_module_prefix=config.user_module_prefix,
        render_as_batch=config.render_as_batch,
        process_revision_directives=writer,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a
    connection with the context.

    Raises:
        RuntimeError: If the engine cannot be created from the config.
    """
    configuration = config.get_section(config.config_ini_section) or {}
    configuration["sqlalchemy.url"] = config.db_url

    connectable = cast(
        "AsyncEngine",
        config.engine
        or async_engine_from_config(
            configuration,
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
            future=True,
        ),
    )
    if connectable is None:  # pyright: ignore[reportUnnecessaryComparison]
        msg = "Could not get engine from config.  Please ensure your `alembic.ini` according to the official Alembic documentation."
        raise RuntimeError(
            msg,
        )

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


if context.is_offline_mode():
    run_migrations_offline()
else:
    asyncio.run(run_migrations_online())



import sys
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union, cast

if TYPE_CHECKING:
    from alembic.migration import MigrationContext
    from alembic.operations.ops import MigrationScript, UpgradeOps
    from click import Group

    from advanced_alchemy.config import SQLAlchemyAsyncConfig, SQLAlchemySyncConfig

__all__ = ("add_migration_commands", "get_alchemy_group")


def get_alchemy_group() -> "Group":
    """Get the Advanced Alchemy CLI group.

    Raises:
        MissingDependencyError: If the `click` package is not installed.

    Returns:
        The Advanced Alchemy CLI group.
    """
    from advanced_alchemy.exceptions import MissingDependencyError
    from advanced_alchemy.utils.cli_tools import click, group

    if click is None:  # pragma: no cover - defensive guard
        raise MissingDependencyError(package="click", install_package="cli")

    @group(name="alchemy")  # pyright: ignore
    @click.option(
        "--config",
        help="Dotted path to SQLAlchemy config(s) (e.g. 'myapp.config.alchemy_configs')",
        required=True,
        type=str,
    )
    @click.pass_context
    def alchemy_group(ctx: "click.Context", config: str) -> None:
        """Advanced Alchemy CLI commands."""
        from pathlib import Path

        from rich import get_console

        from advanced_alchemy.utils import module_loader

        console = get_console()
        ctx.ensure_object(dict)

        # Add current working directory to sys.path to allow loading local config modules
        cwd = str(Path.cwd())
        if cwd not in sys.path:
            sys.path.insert(0, cwd)

        try:
            config_instance = module_loader.import_string(config)
            if isinstance(config_instance, Sequence):
                ctx.obj["configs"] = config_instance
            else:
                ctx.obj["configs"] = [config_instance]
        except ImportError as e:
            console.print(f"[red]Error loading config: {e}[/]")
            ctx.exit(1)
        finally:
            # Clean up: remove the cwd from sys.path if we added it
            if cwd in sys.path and sys.path[0] == cwd:
                sys.path.remove(cwd)

    return alchemy_group


def add_migration_commands(database_group: Optional["Group"] = None) -> "Group":  # noqa: C901, PLR0915
    """Add migration commands to the database group.

    Args:
        database_group: The database group to add the commands to.

    Raises:
        MissingDependencyError: If the `click` package is not installed.

    Returns:
        The database group with the migration commands added.
    """
    from rich import get_console

    from advanced_alchemy.utils.cli_tools import click

    console = get_console()

    if database_group is None:
        database_group = get_alchemy_group()

    bind_key_option = click.option(
        "--bind-key",
        help="Specify which SQLAlchemy config to use by bind key",
        type=str,
        default=None,
    )
    verbose_option = click.option(
        "--verbose",
        help="Enable verbose output.",
        type=bool,
        default=False,
        is_flag=True,
    )
    no_prompt_option = click.option(
        "--no-prompt",
        help="Do not prompt for confirmation before executing the command.",
        type=bool,
        default=False,
        required=False,
        show_default=True,
        is_flag=True,
    )

    def get_config_by_bind_key(
        ctx: "click.Context", bind_key: Optional[str]
    ) -> "Union[SQLAlchemyAsyncConfig, SQLAlchemySyncConfig]":
        """Get the SQLAlchemy config for the specified bind key.

        Args:
            ctx: The click context.
            bind_key: The bind key to get the config for.

        Returns:
            The SQLAlchemy config for the specified bind key.
        """
        configs = ctx.obj["configs"]
        if bind_key is None:
            return cast("Union[SQLAlchemyAsyncConfig, SQLAlchemySyncConfig]", configs[0])

        for config in configs:
            if config.bind_key == bind_key:
                return cast("Union[SQLAlchemyAsyncConfig, SQLAlchemySyncConfig]", config)

        console.print(f"[red]No config found for bind key: {bind_key}[/]")
        sys.exit(1)

    @database_group.command(
        name="show-current-revision",
        help="Shows the current revision for the database.",
    )
    @bind_key_option
    @verbose_option
    def show_database_revision(bind_key: Optional[str], verbose: bool) -> None:  # pyright: ignore[reportUnusedFunction]
        """Show current database revision."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Listing current revision[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.current(verbose=verbose)

    @database_group.command(
        name="downgrade",
        help="Downgrade database to a specific revision.",
    )
    @bind_key_option
    @click.option("--sql", type=bool, help="Generate SQL output for offline migrations.", default=False, is_flag=True)
    @click.option(
        "--tag",
        help="an arbitrary 'tag' that can be intercepted by custom env.py scripts via the .EnvironmentContext.get_tag_argument method.",
        type=str,
        default=None,
    )
    @no_prompt_option
    @click.argument(
        "revision",
        type=str,
        default="-1",
    )
    def downgrade_database(  # pyright: ignore[reportUnusedFunction]
        bind_key: Optional[str], revision: str, sql: bool, tag: Optional[str], no_prompt: bool
    ) -> None:
        """Downgrade the database to the latest revision."""
        from rich.prompt import Confirm

        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Starting database downgrade process[/]", align="left")
        input_confirmed = (
            True
            if no_prompt
            else Confirm.ask(f"Are you sure you want to downgrade the database to the `{revision}` revision?")
        )
        if input_confirmed:
            sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
            alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
            alembic_commands.downgrade(revision=revision, sql=sql, tag=tag)

    @database_group.command(
        name="upgrade",
        help="Upgrade database to a specific revision.",
    )
    @bind_key_option
    @click.option("--sql", type=bool, help="Generate SQL output for offline migrations.", default=False, is_flag=True)
    @click.option(
        "--tag",
        help="an arbitrary 'tag' that can be intercepted by custom env.py scripts via the .EnvironmentContext.get_tag_argument method.",
        type=str,
        default=None,
    )
    @no_prompt_option
    @click.argument(
        "revision",
        type=str,
        default="head",
    )
    def upgrade_database(  # pyright: ignore[reportUnusedFunction]
        bind_key: Optional[str], revision: str, sql: bool, tag: Optional[str], no_prompt: bool
    ) -> None:
        """Upgrade the database to the latest revision."""
        from rich.prompt import Confirm

        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Starting database upgrade process[/]", align="left")
        input_confirmed = (
            True
            if no_prompt
            else Confirm.ask(f"[bold]Are you sure you want migrate the database to the `{revision}` revision?[/]")
        )
        if input_confirmed:
            sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
            alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
            alembic_commands.upgrade(revision=revision, sql=sql, tag=tag)

    @database_group.command(
        help="Stamp the revision table with the given revision; don't run any migrations",
    )
    @click.argument("revision", type=str)
    @bind_key_option
    @click.option("--sql", is_flag=True, default=False, help="Generate SQL output for offline migrations")
    @click.option(
        "--tag", type=str, default=None, help="Arbitrary 'tag' that can be intercepted by custom env.py scripts"
    )
    @click.option("--purge", is_flag=True, default=False, help="Delete all entries in version table before stamping")
    def stamp(bind_key: Optional[str], revision: str, sql: bool, tag: Optional[str], purge: bool) -> None:  # pyright: ignore[reportUnusedFunction]
        """Stamp the revision table with the given revision."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Stamping revision table[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.stamp(revision=revision, sql=sql, tag=tag, purge=purge)

    @database_group.command(
        name="check",
        help="Check if the target database is up to date",
    )
    @bind_key_option
    def check_revision(bind_key: Optional[str]) -> None:  # pyright: ignore[reportUnusedFunction]
        """Check for pending upgrade operations."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Checking for pending migrations[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.check()

    @database_group.command(
        name="edit",
        help="Edit a revision file using $EDITOR",
    )
    @click.argument("revision", type=str)
    @bind_key_option
    def edit_revision(bind_key: Optional[str], revision: str) -> None:  # pyright: ignore[reportUnusedFunction]
        """Edit revision script with system editor."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule(f"[yellow]Opening revision {revision} in editor[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.edit(revision=revision)

    @database_group.command(
        name="ensure-version",
        help="Create the alembic version table if it doesn't exist",
    )
    @bind_key_option
    @click.option("--sql", is_flag=True, default=False, help="Generate SQL output instead of executing")
    def ensure_version_table(bind_key: Optional[str], sql: bool) -> None:  # pyright: ignore[reportUnusedFunction]
        """Ensure alembic version table exists."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Ensuring version table exists[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.ensure_version(sql=sql)

    @database_group.command(
        name="heads",
        help="Show current available heads in the script directory",
    )
    @bind_key_option
    @verbose_option
    @click.option(
        "--resolve-dependencies",
        is_flag=True,
        default=False,
        help="Resolve dependencies between heads",
    )
    def show_heads(bind_key: Optional[str], verbose: bool, resolve_dependencies: bool) -> None:  # pyright: ignore[reportUnusedFunction]
        """Show current heads."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Showing current heads[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.heads(verbose=verbose, resolve_dependencies=resolve_dependencies)

    @database_group.command(
        name="history",
        help="List changeset scripts in chronological order",
    )
    @bind_key_option
    @verbose_option
    @click.option(
        "--rev-range",
        type=str,
        default=None,
        help="Revision range (e.g., 'base:head', 'abc:def')",
    )
    @click.option(
        "--indicate-current",
        is_flag=True,
        default=False,
        help="Indicate the current revision",
    )
    def show_history(  # pyright: ignore[reportUnusedFunction]
        bind_key: Optional[str],
        verbose: bool,
        rev_range: Optional[str],
        indicate_current: bool,
    ) -> None:
        """Show revision history."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Showing revision history[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.history(
            rev_range=rev_range,
            verbose=verbose,
            indicate_current=indicate_current,
        )

    @database_group.command(
        name="merge",
        help="Merge two revisions together, creating a new migration file",
    )
    @click.argument("revisions", type=str)
    @bind_key_option
    @click.option("-m", "--message", type=str, default=None, help="Merge message")
    @click.option("--branch-label", type=str, default=None, help="Branch label for merge revision")
    @click.option("--rev-id", type=str, default=None, help="Specify custom revision ID")
    @no_prompt_option
    def merge_revisions(  # pyright: ignore[reportUnusedFunction]
        bind_key: Optional[str],
        revisions: str,
        message: Optional[str],
        branch_label: Optional[str],
        rev_id: Optional[str],
        no_prompt: bool,
    ) -> None:
        """Merge revisions (resolves multiple heads)."""
        from rich.prompt import Prompt

        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Merging revisions[/]", align="left")

        if message is None:
            message = "merge revisions" if no_prompt else Prompt.ask("Enter merge message")

        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.merge(
            revisions=revisions,
            message=message,
            branch_label=branch_label,
            rev_id=rev_id,
        )

    @database_group.command(
        name="show",
        help="Show the revision denoted by the given symbol",
    )
    @click.argument("revision", type=str)
    @bind_key_option
    def show_revision(bind_key: Optional[str], revision: str) -> None:  # pyright: ignore[reportUnusedFunction]
        """Show details of a specific revision."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule(f"[yellow]Showing revision {revision}[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.show(rev=revision)

    @database_group.command(
        name="branches",
        help="Show current branch points in the migration history",
    )
    @bind_key_option
    @verbose_option
    def show_branches(bind_key: Optional[str], verbose: bool) -> None:  # pyright: ignore[reportUnusedFunction]
        """Show branch points."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Showing branch points[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.branches(verbose=verbose)

    @database_group.command(
        name="list-templates",
        help="List available Alembic migration templates",
    )
    @bind_key_option
    def list_init_templates(bind_key: Optional[str]) -> None:  # pyright: ignore[reportUnusedFunction]
        """List available initialization templates."""
        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Available templates[/]", align="left")
        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.list_templates()

    @database_group.command(
        name="init",
        help="Initialize migrations for the project.",
    )
    @bind_key_option
    @click.argument(
        "directory",
        default=None,
        required=False,
    )
    @click.option("--multidb", is_flag=True, default=False, help="Support multiple databases")
    @click.option("--package", is_flag=True, default=True, help="Create `__init__.py` for created folder")
    @no_prompt_option
    def init_alembic(  # pyright: ignore[reportUnusedFunction]
        bind_key: Optional[str], directory: Optional[str], multidb: bool, package: bool, no_prompt: bool
    ) -> None:
        """Initialize the database migrations."""
        from rich.prompt import Confirm

        from advanced_alchemy.alembic.commands import AlembicCommands

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Initializing database migrations.", align="left")
        input_confirmed = (
            True if no_prompt else Confirm.ask("[bold]Are you sure you want initialize migrations for the project?[/]")
        )
        if input_confirmed:
            configs = [get_config_by_bind_key(ctx, bind_key)] if bind_key is not None else ctx.obj["configs"]
            for config in configs:
                directory = config.alembic_config.script_location if directory is None else directory
                alembic_commands = AlembicCommands(sqlalchemy_config=config)
                alembic_commands.init(directory=cast("str", directory), multidb=multidb, package=package)

    @database_group.command(
        name="make-migrations",
        help="Create a new migration revision.",
    )
    @bind_key_option
    @click.option("-m", "--message", default=None, help="Revision message")
    @click.option(
        "--autogenerate/--no-autogenerate", default=True, help="Automatically populate revision with detected changes"
    )
    @click.option("--sql", is_flag=True, default=False, help="Export to `.sql` instead of writing to the database.")
    @click.option("--head", default="head", help="Specify head revision to use as base for new revision.")
    @click.option(
        "--splice", is_flag=True, default=False, help='Allow a non-head revision as the "head" to splice onto'
    )
    @click.option("--branch-label", default=None, help="Specify a branch label to apply to the new revision")
    @click.option("--version-path", default=None, help="Specify specific path from config for version file")
    @click.option("--rev-id", default=None, help="Specify a ID to use for revision.")
    @no_prompt_option
    def create_revision(  # pyright: ignore[reportUnusedFunction]
        bind_key: Optional[str],
        message: Optional[str],
        autogenerate: bool,
        sql: bool,
        head: str,
        splice: bool,
        branch_label: Optional[str],
        version_path: Optional[str],
        rev_id: Optional[str],
        no_prompt: bool,
    ) -> None:
        """Create a new database revision."""
        from rich.prompt import Prompt

        from advanced_alchemy.alembic.commands import AlembicCommands

        def process_revision_directives(
            context: "MigrationContext",  # noqa: ARG001
            revision: tuple[str],  # noqa: ARG001
            directives: list["MigrationScript"],
        ) -> None:
            """Handle revision directives."""
            if autogenerate and cast("UpgradeOps", directives[0].upgrade_ops).is_empty():
                console.rule(
                    "[magenta]The generation of a migration file is being skipped because it would result in an empty file.",
                    style="magenta",
                    align="left",
                )
                console.rule(
                    "[magenta]More information can be found here. https://alembic.sqlalchemy.org/en/latest/autogenerate.html#what-does-autogenerate-detect-and-what-does-it-not-detect",
                    style="magenta",
                    align="left",
                )
                console.rule(
                    "[magenta]If you intend to create an empty migration file, use the --no-autogenerate option.",
                    style="magenta",
                    align="left",
                )
                directives.clear()

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Starting database upgrade process[/]", align="left")
        if message is None:
            message = "autogenerated" if no_prompt else Prompt.ask("Please enter a message describing this revision")

        sqlalchemy_config = get_config_by_bind_key(ctx, bind_key)
        alembic_commands = AlembicCommands(sqlalchemy_config=sqlalchemy_config)
        alembic_commands.revision(
            message=message,
            autogenerate=autogenerate,
            sql=sql,
            head=head,
            splice=splice,
            branch_label=branch_label,
            version_path=version_path,
            rev_id=rev_id,
            process_revision_directives=process_revision_directives,  # type: ignore[arg-type]
        )

    @database_group.command(name="drop-all", help="Drop all tables from the database.")
    @bind_key_option
    @no_prompt_option
    def drop_all(bind_key: Optional[str], no_prompt: bool) -> None:  # pyright: ignore[reportUnusedFunction]
        """Drop all tables from the database."""
        from anyio import run
        from rich.prompt import Confirm

        from advanced_alchemy.alembic.utils import drop_all
        from advanced_alchemy.base import metadata_registry

        ctx = cast("click.Context", click.get_current_context())
        console.rule("[yellow]Dropping all tables from the database[/]", align="left")
        input_confirmed = no_prompt or Confirm.ask(
            "[bold red]Are you sure you want to drop all tables from the database?"
        )

        async def _drop_all(
            configs: "Sequence[Union[SQLAlchemyAsyncConfig, SQLAlchemySyncConfig]]",
        ) -> None:
            for config in configs:
                engine = config.get_engine()
                await drop_all(engine, config.alembic_config.version_table_name, metadata_registry.get(config.bind_key))

        if input_confirmed:
            configs = [get_config_by_bind_key(ctx, bind_key)] if bind_key is not None else ctx.obj["configs"]
            run(_drop_all, configs)

    @database_group.command(name="dump-data", help="Dump specified tables from the database to JSON files.")
    @bind_key_option
    @click.option(
        "--table",
        "table_names",
        help="Name of the table to dump. Multiple tables can be specified. Use '*' to dump all tables.",
        type=str,
        required=True,
        multiple=True,
    )
    @click.option(
        "--dir",
        "dump_dir",
        help="Directory to save the JSON files. Defaults to WORKDIR/fixtures",
        type=click.Path(path_type=Path),
        default=Path.cwd() / "fixtures",
        required=False,
    )
    def dump_table_data(bind_key: Optional[str], table_names: tuple[str, ...], dump_dir: Path) -> None:  # pyright: ignore[reportUnusedFunction]
        """Dump table data to JSON files."""
        from anyio import run
        from rich.prompt import Confirm

        from advanced_alchemy.alembic.utils import dump_tables
        from advanced_alchemy.base import metadata_registry, orm_registry

        ctx = cast("click.Context", click.get_current_context())
        all_tables = "*" in table_names

        if all_tables and not Confirm.ask(
            "[yellow bold]You have specified '*'. Are you sure you want to dump all tables from the database?",
        ):
            return console.rule("[red bold]No data was dumped.", style="red", align="left")

        async def _dump_tables() -> None:
            configs = [get_config_by_bind_key(ctx, bind_key)] if bind_key is not None else ctx.obj["configs"]
            for config in configs:
                target_tables = set(metadata_registry.get(config.bind_key).tables)

                if not all_tables:
                    for table_name in set(table_names) - target_tables:
                        console.rule(
                            f"[red bold]Skipping table '{table_name}' because it is not available in the default registry",
                            style="red",
                            align="left",
                        )
                    target_tables.intersection_update(table_names)
                else:
                    console.rule("[yellow bold]Dumping all tables", style="yellow", align="left")

                models = [
                    mapper.class_ for mapper in orm_registry.mappers if mapper.class_.__table__.name in target_tables
                ]
                await dump_tables(dump_dir, config.get_session(), models)
                console.rule("[green bold]Data dump complete", align="left")

        return run(_dump_tables)

    return database_group



"""Unit tests for replica selectors.

Tests different strategies for selecting read replicas.
"""

from unittest.mock import MagicMock

import pytest
from sqlalchemy import Engine

from advanced_alchemy.routing.selectors import RandomSelector, RoundRobinSelector


@pytest.fixture
def mock_engines() -> list[Engine]:
    """Create mock engines for testing."""
    return [MagicMock(name=f"engine_{i}") for i in range(3)]


def test_round_robin_selector_initialization(mock_engines: list[Engine]) -> None:
    """Test RoundRobinSelector initialization."""
    selector = RoundRobinSelector(mock_engines)

    assert selector.has_replicas() is True
    assert len(selector.replicas) == 3


def test_round_robin_selector_empty_initialization() -> None:
    """Test RoundRobinSelector initialization with no replicas."""
    selector: RoundRobinSelector[Engine] = RoundRobinSelector([])

    assert selector.has_replicas() is False
    assert len(selector.replicas) == 0


def test_round_robin_selector_cycles_through_replicas(mock_engines: list[Engine]) -> None:
    """Test that RoundRobinSelector cycles through replicas in order."""
    selector = RoundRobinSelector(mock_engines)

    assert selector.next() is mock_engines[0]
    assert selector.next() is mock_engines[1]
    assert selector.next() is mock_engines[2]

    assert selector.next() is mock_engines[0]
    assert selector.next() is mock_engines[1]
    assert selector.next() is mock_engines[2]


def test_round_robin_selector_single_replica() -> None:
    """Test RoundRobinSelector with a single replica."""
    engine = MagicMock(name="single_engine")
    selector = RoundRobinSelector([engine])

    assert selector.next() is engine
    assert selector.next() is engine
    assert selector.next() is engine


def test_round_robin_selector_no_replicas_raises() -> None:
    """Test that RoundRobinSelector raises RuntimeError when no replicas configured."""
    selector: RoundRobinSelector[Engine] = RoundRobinSelector([])

    with pytest.raises(RuntimeError, match="No engines configured for round-robin selection"):
        selector.next()


def test_round_robin_selector_thread_safe(mock_engines: list[Engine]) -> None:
    """Test that RoundRobinSelector is thread-safe (uses lock)."""
    import threading

    selector = RoundRobinSelector(mock_engines)
    results: list[Engine] = []

    def worker() -> None:
        results.extend([selector.next() for _ in range(10)])

    threads = [threading.Thread(target=worker) for _ in range(5)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    assert len(results) == 50
    assert all(engine in mock_engines for engine in results)


def test_random_selector_initialization(mock_engines: list[Engine]) -> None:
    """Test RandomSelector initialization."""
    selector = RandomSelector(mock_engines)

    assert selector.has_replicas() is True
    assert len(selector.replicas) == 3


def test_random_selector_empty_initialization() -> None:
    """Test RandomSelector initialization with no replicas."""
    selector: RandomSelector[Engine] = RandomSelector([])

    assert selector.has_replicas() is False
    assert len(selector.replicas) == 0


def test_random_selector_returns_valid_replica(mock_engines: list[Engine]) -> None:
    """Test that RandomSelector returns valid replicas."""
    selector = RandomSelector(mock_engines)

    selections = [selector.next() for _ in range(100)]

    assert all(engine in mock_engines for engine in selections)


def test_random_selector_distributes_randomly(mock_engines: list[Engine]) -> None:
    """Test that RandomSelector distributes selections across replicas."""
    selector = RandomSelector(mock_engines)

    selections = [selector.next() for _ in range(1000)]

    counts = {engine: selections.count(engine) for engine in mock_engines}

    assert all(count > 0 for count in counts.values())

    for count in counts.values():
        assert 200 < count < 450, f"Distribution not random enough: {counts}"


def test_random_selector_single_replica() -> None:
    """Test RandomSelector with a single replica."""
    engine = MagicMock(name="single_engine")
    selector = RandomSelector([engine])

    assert selector.next() is engine
    assert selector.next() is engine
    assert selector.next() is engine


def test_random_selector_no_replicas_raises() -> None:
    """Test that RandomSelector raises RuntimeError when no replicas configured."""
    selector: RandomSelector[Engine] = RandomSelector([])

    with pytest.raises(RuntimeError, match="No engines configured for random selection"):
        selector.next()


def test_replica_selector_replicas_property(mock_engines: list[Engine]) -> None:
    """Test that replica selector exposes replicas property."""
    selector = RoundRobinSelector(mock_engines)

    assert selector.replicas == mock_engines


def test_has_replicas_returns_false_for_empty_list() -> None:
    """Test has_replicas returns False for empty replica list."""
    selector: RoundRobinSelector[Engine] = RoundRobinSelector([])

    assert selector.has_replicas() is False


def test_has_replicas_returns_true_for_non_empty_list(mock_engines: list[Engine]) -> None:
    """Test has_replicas returns True for non-empty replica list."""
    selector = RoundRobinSelector(mock_engines)

    assert selector.has_replicas() is True



from __future__ import annotations

from typing import TYPE_CHECKING

from tools.sphinx_ext import changelog, missing_references

if TYPE_CHECKING:
    from sphinx.application import Sphinx


def setup(app: Sphinx) -> dict[str, bool]:
    ext_config = {}
    ext_config.update(missing_references.setup(app))
    ext_config.update(changelog.setup(app))  # type: ignore[arg-type]

    return ext_config  # pyright: ignore



from typing import TypedDict

from sqlalchemy.orm import Mapped, MappedAsDataclass, declarative_mixin, declared_attr, mapped_column
from sqlalchemy.sql.schema import _InsertSentinelColumnDefault  # pyright: ignore [reportPrivateUsage]
from typing_extensions import NotRequired


class SentinelKwargs(TypedDict):
    init: NotRequired[bool]


@declarative_mixin
class SentinelMixin:
    """Mixin to add a sentinel column for SQLAlchemy models."""

    __abstract__ = True
    _sentinel_kwargs: SentinelKwargs = {}

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if issubclass(cls, MappedAsDataclass):
            cls._sentinel_kwargs["init"] = False

    @declared_attr
    def _sentinel(cls) -> Mapped[int]:
        return mapped_column(
            name="sa_orm_sentinel",
            insert_default=_InsertSentinelColumnDefault(),
            _omit_from_statements=True,
            insert_sentinel=True,
            use_existing_column=True,
            nullable=True,
            sort_order=3001,
            **cls._sentinel_kwargs,
        )



from advanced_alchemy.extensions.litestar.plugins.init.config.asyncio import SQLAlchemyAsyncConfig
from advanced_alchemy.extensions.litestar.plugins.init.config.engine import EngineConfig
from advanced_alchemy.extensions.litestar.plugins.init.config.sync import SQLAlchemySyncConfig

__all__ = (
    "EngineConfig",
    "SQLAlchemyAsyncConfig",
    "SQLAlchemySyncConfig",
)



"""Context variables and context managers for read/write routing.

This module provides the context-based state management for routing decisions,
including the sticky-to-primary behavior after writes.
"""

from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import Optional

__all__ = (
    "bind_group_var",
    "force_primary_var",
    "primary_context",
    "replica_context",
    "reset_routing_context",
    "stick_to_primary_var",
    "use_bind_group",
)


stick_to_primary_var: ContextVar[bool] = ContextVar("stick_to_primary", default=False)
"""Context variable tracking if we should stick to primary after a write.

When ``True``, all operations (including reads) will use the primary database
until the context is reset (typically after commit/rollback).
"""

force_primary_var: ContextVar[bool] = ContextVar("force_primary", default=False)
"""Context variable for explicitly forcing all operations to primary.

When ``True``, all operations will use the primary database regardless
of operation type or stickiness state.
"""

bind_group_var: ContextVar[Optional[str]] = ContextVar("bind_group", default=None)
"""Context variable for explicitly selecting a bind group.

When set, this overrides the automatic routing logic to use the specified
engine group (e.g., "analytics", "reader").
"""


@contextmanager
def primary_context() -> Generator[None, None, None]:
    """Force all operations to use primary within this context.

    Use this context manager when you need to ensure all database operations
    (including reads) go to the primary database.

    Example:
        Force a specific query to use the primary database::

            from advanced_alchemy.routing import primary_context

            with primary_context():
                user = await repo.get(user_id)
                orders = await order_repo.get_many()

    Yields:
        None
    """
    token: Token[bool] = force_primary_var.set(True)
    try:
        yield
    finally:
        force_primary_var.reset(token)


@contextmanager
def replica_context() -> Generator[None, None, None]:
    """Force read operations to use replicas (temporarily disable stickiness).

    Use this context manager when you want to explicitly allow reads to go
    to replicas, even if a previous write has set the sticky-to-primary state.

    .. warning::
        Use with caution! This can lead to read-after-write inconsistency
        if you're reading data that was recently written.

    Example:
        Allow reads to use replicas after a write::

            from advanced_alchemy.routing import replica_context

            await repo.add(user)

            user = await repo.get(user_id)

            with replica_context():
                users = await repo.get_many()

    Yields:
        None
    """
    stick_token: Token[bool] = stick_to_primary_var.set(False)
    force_token: Token[bool] = force_primary_var.set(False)
    try:
        yield
    finally:
        stick_to_primary_var.reset(stick_token)
        force_primary_var.reset(force_token)


@contextmanager
def use_bind_group(name: str) -> Generator[None, None, None]:
    """Force operations to use a specific bind group.

    Use this context manager to route operations to a specific group of engines,
    such as "analytics" or "reporting".

    Example:
        Route a query to the analytics database::

            from advanced_alchemy.routing import use_bind_group

            with use_bind_group("analytics"):
                data = await repo.get_many()

    Args:
        name: Name of the bind group to use.

    Yields:
        None
    """
    token: Token[Optional[str]] = bind_group_var.set(name)
    try:
        yield
    finally:
        bind_group_var.reset(token)


def reset_routing_context() -> None:
    """Reset all routing context variables to their defaults.

    This is typically called after a commit or rollback to allow
    subsequent reads to use replicas again.

    Example:
        Manual reset after transaction::

            from advanced_alchemy.routing import reset_routing_context

            await session.commit()
            reset_routing_context()
    """
    stick_to_primary_var.set(False)
    force_primary_var.set(False)
    bind_group_var.set(None)


def set_sticky_primary() -> None:
    """Set the sticky-to-primary flag.

    This is called internally after write operations to ensure
    subsequent reads use the primary database.
    """
    stick_to_primary_var.set(True)


def should_use_primary() -> bool:
    """Check if we should route to the primary database.

    Returns:
        ``True`` if routing should use primary (due to force or stickiness).
    """
    return force_primary_var.get() or stick_to_primary_var.get()



"""SQLAlchemy custom types for use with the ORM."""

from advanced_alchemy.types import encrypted_string, file_object, password_hash, totp
from advanced_alchemy.types.boolean import Bool
from advanced_alchemy.types.datetime import DateTimeUTC
from advanced_alchemy.types.encrypted_string import (
    EncryptedString,
    EncryptedText,
    EncryptionBackend,
    FernetBackend,
)
from advanced_alchemy.types.file_object import (
    FileObject,
    FileObjectList,
    StorageBackend,
    StorageBackendT,
    StorageRegistry,
    StoredObject,
    storages,
)
from advanced_alchemy.types.guid import GUID, NANOID_INSTALLED, UUID_UTILS_INSTALLED
from advanced_alchemy.types.identity import BigIntIdentity
from advanced_alchemy.types.json import ORA_JSONB, JsonB
from advanced_alchemy.types.mutables import MutableList
from advanced_alchemy.types.password_hash.base import HashedPassword, PasswordHash
from advanced_alchemy.types.password_hash.one_time_code import (
    HashedOneTimeCode,
    OneTimeCode,
    generate_one_time_code,
)
from advanced_alchemy.types.totp import TOTPProvider, TOTPSecret, generate_totp_secret
from advanced_alchemy.types.vector import Vector

__all__ = (
    "GUID",
    "NANOID_INSTALLED",
    "ORA_JSONB",
    "UUID_UTILS_INSTALLED",
    "BigIntIdentity",
    "Bool",
    "DateTimeUTC",
    "EncryptedString",
    "EncryptedText",
    "EncryptionBackend",
    "FernetBackend",
    "FileObject",
    "FileObjectList",
    "HashedOneTimeCode",
    "HashedPassword",
    "JsonB",
    "MutableList",
    "OneTimeCode",
    "PasswordHash",
    "StorageBackend",
    "StorageBackendT",
    "StorageRegistry",
    "StoredObject",
    "TOTPProvider",
    "TOTPSecret",
    "Vector",
    "encrypted_string",
    "file_object",
    "generate_one_time_code",
    "generate_totp_secret",
    "password_hash",
    "storages",
    "totp",
)



"""Integration tests for read/write routing functionality.

These tests verify that the routing module correctly routes read operations to
replicas and write operations to the primary database with real sessions.
"""

import asyncio
from collections.abc import Generator
from pathlib import Path
from uuid import uuid4

import pytest
from sqlalchemy import Engine, String, create_engine, select
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, Session, mapped_column

from advanced_alchemy.config.routing import RoutingConfig, RoutingStrategy
from advanced_alchemy.routing import (
    RandomSelector,
    RoundRobinSelector,
    RoutingAsyncSession,
    RoutingAsyncSessionMaker,
    RoutingSyncSession,
    RoutingSyncSessionMaker,
    primary_context,
    replica_context,
    reset_routing_context,
)

pytestmark = [
    pytest.mark.integration,
    pytest.mark.xdist_group("routing"),
]


class RoutingTestBase(DeclarativeBase):
    """Base class for routing test models."""

    pass


class User(RoutingTestBase):
    """Simple user model for routing tests."""

    __tablename__ = "routing_users"

    id: Mapped[str] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid4()))
    name: Mapped[str] = mapped_column(String(100))


@pytest.fixture()
def routing_test_db_paths(tmp_path: Path) -> dict[str, Path]:
    """Create temporary database file paths for primary and replicas."""
    return {
        "primary": tmp_path / "primary.db",
        "replica1": tmp_path / "replica1.db",
        "replica2": tmp_path / "replica2.db",
    }


@pytest.fixture()
def sync_routing_engines(routing_test_db_paths: dict[str, Path]) -> Generator[dict[str, Engine], None, None]:
    """Create sync engines for primary and replicas."""
    engines = {
        "primary": create_engine(f"sqlite:///{routing_test_db_paths['primary']}"),
        "replica1": create_engine(f"sqlite:///{routing_test_db_paths['replica1']}"),
        "replica2": create_engine(f"sqlite:///{routing_test_db_paths['replica2']}"),
    }

    for engine in engines.values():
        RoutingTestBase.metadata.create_all(engine)

    yield engines

    for engine in engines.values():
        engine.dispose()


@pytest.fixture()
def async_routing_engines(
    routing_test_db_paths: dict[str, Path],
) -> Generator[dict[str, AsyncEngine], None, None]:
    """Create async engines for primary and replicas."""
    engines = {
        "primary": create_async_engine(f"sqlite+aiosqlite:///{routing_test_db_paths['primary']}"),
        "replica1": create_async_engine(f"sqlite+aiosqlite:///{routing_test_db_paths['replica1']}"),
        "replica2": create_async_engine(f"sqlite+aiosqlite:///{routing_test_db_paths['replica2']}"),
    }

    yield engines


@pytest.fixture()
def routing_config() -> RoutingConfig:
    """Create a basic routing configuration."""
    return RoutingConfig(
        primary_connection_string="sqlite:///primary.db",
        read_replicas=["sqlite:///replica1.db", "sqlite:///replica2.db"],
        routing_strategy=RoutingStrategy.ROUND_ROBIN,
        sticky_after_write=True,
        reset_stickiness_on_commit=True,
    )


@pytest.fixture()
def sync_session_maker(
    routing_test_db_paths: dict[str, Path],
) -> Generator[RoutingSyncSessionMaker, None, None]:
    """Create a sync routing session maker with real databases."""
    config = RoutingConfig(
        primary_connection_string=f"sqlite:///{routing_test_db_paths['primary']}",
        read_replicas=[
            f"sqlite:///{routing_test_db_paths['replica1']}",
            f"sqlite:///{routing_test_db_paths['replica2']}",
        ],
        routing_strategy=RoutingStrategy.ROUND_ROBIN,
        sticky_after_write=True,
    )

    maker = RoutingSyncSessionMaker(routing_config=config)

    RoutingTestBase.metadata.create_all(maker.primary_engine)
    for engine in maker.replica_engines:
        RoutingTestBase.metadata.create_all(engine)

    yield maker

    maker.close_all()


@pytest.fixture()
def async_session_maker(
    routing_test_db_paths: dict[str, Path],
) -> Generator[RoutingAsyncSessionMaker, None, None]:
    """Create an async routing session maker with real databases."""
    config = RoutingConfig(
        primary_connection_string=f"sqlite+aiosqlite:///{routing_test_db_paths['primary']}",
        read_replicas=[
            f"sqlite+aiosqlite:///{routing_test_db_paths['replica1']}",
            f"sqlite+aiosqlite:///{routing_test_db_paths['replica2']}",
        ],
        routing_strategy=RoutingStrategy.ROUND_ROBIN,
        sticky_after_write=True,
    )

    maker = RoutingAsyncSessionMaker(routing_config=config)

    yield maker


class TestRoutingSyncSession:
    """Integration tests for sync routing sessions."""

    def test_write_goes_to_primary(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that INSERT statements are routed to primary."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            user = User(name="Test User")
            session.add(user)
            session.commit()

            with Session(sync_session_maker.primary_engine) as primary_session:
                result = primary_session.execute(select(User).where(User.name == "Test User")).scalar_one_or_none()
                assert result is not None
                assert result.name == "Test User"

        finally:
            session.close()
            reset_routing_context()

    def test_read_after_write_stickiness(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that reads stick to primary after a write."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            with Session(sync_session_maker.primary_engine) as primary_session:
                primary_session.add(User(id="primary-user", name="Primary User"))
                primary_session.commit()

            session.add(User(name="Trigger Write"))
            session.flush()

            stmt = select(User).where(User.id == "primary-user")
            result = session.execute(stmt).scalar_one_or_none()
            assert result is not None
            assert result.name == "Primary User"

        finally:
            session.rollback()
            session.close()
            reset_routing_context()

    def test_commit_resets_stickiness(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that commit resets stickiness when configured."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            user = User(name="Test User")
            session.add(user)
            session.commit()

            from advanced_alchemy.routing.context import stick_to_primary_var

            assert not stick_to_primary_var.get()

        finally:
            session.close()
            reset_routing_context()

    def test_rollback_resets_stickiness(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that rollback always resets stickiness."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            session.add(User(name="Test User"))
            session.flush()

            from advanced_alchemy.routing.context import stick_to_primary_var

            assert stick_to_primary_var.get()

            session.rollback()

            assert not stick_to_primary_var.get()

        finally:
            session.close()
            reset_routing_context()

    def test_primary_context_forces_primary(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that primary_context() forces operations to primary."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            with Session(sync_session_maker.primary_engine) as primary_session:
                primary_session.add(User(id="primary-only", name="Primary Only"))
                primary_session.commit()

            with primary_context():
                result = session.execute(select(User).where(User.id == "primary-only")).scalar_one_or_none()
                assert result is not None
                assert result.name == "Primary Only"

        finally:
            session.close()
            reset_routing_context()

    def test_replica_context_allows_replicas_after_write(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that replica_context() allows replicas even after writes."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            session.add(User(name="Trigger Write"))
            session.flush()

            from advanced_alchemy.routing.context import stick_to_primary_var

            assert stick_to_primary_var.get()

            with replica_context():
                from advanced_alchemy.routing.context import force_primary_var

                assert not force_primary_var.get()

        finally:
            session.rollback()
            session.close()
            reset_routing_context()


class TestRoutingAsyncSession:
    """Integration tests for async routing sessions."""

    @pytest.mark.asyncio
    async def test_async_write_goes_to_primary(self, async_session_maker: RoutingAsyncSessionMaker) -> None:
        """Test that async INSERT statements are routed to primary."""
        reset_routing_context()

        async with async_session_maker.primary_engine.begin() as conn:
            await conn.run_sync(RoutingTestBase.metadata.create_all)
        for replica_engine in async_session_maker.replica_engines:
            async with replica_engine.begin() as conn:
                await conn.run_sync(RoutingTestBase.metadata.create_all)

        session = async_session_maker()

        try:
            user = User(name="Async Test User")
            session.add(user)
            await session.commit()

            async with AsyncSession(async_session_maker.primary_engine) as primary_session:
                result = await primary_session.execute(select(User).where(User.name == "Async Test User"))
                user_result = result.scalar_one_or_none()
                assert user_result is not None
                assert user_result.name == "Async Test User"

        finally:
            await session.close()
            reset_routing_context()

        await async_session_maker.close_all()

    @pytest.mark.asyncio
    async def test_async_primary_context(self, async_session_maker: RoutingAsyncSessionMaker) -> None:
        """Test that primary_context() works with async sessions."""
        reset_routing_context()

        async with async_session_maker.primary_engine.begin() as conn:
            await conn.run_sync(RoutingTestBase.metadata.create_all)

        async with AsyncSession(async_session_maker.primary_engine) as primary_session:
            primary_session.add(User(id="async-primary", name="Async Primary"))
            await primary_session.commit()

        session = async_session_maker()

        try:
            with primary_context():
                result = await session.execute(select(User).where(User.id == "async-primary"))
                user = result.scalar_one_or_none()
                assert user is not None
                assert user.name == "Async Primary"

        finally:
            await session.close()
            reset_routing_context()

        await async_session_maker.close_all()

    @pytest.mark.asyncio
    async def test_async_commit_resets_stickiness(self, async_session_maker: RoutingAsyncSessionMaker) -> None:
        """Test that async commit resets stickiness."""
        reset_routing_context()

        async with async_session_maker.primary_engine.begin() as conn:
            await conn.run_sync(RoutingTestBase.metadata.create_all)

        session = async_session_maker()

        try:
            from advanced_alchemy.routing.context import stick_to_primary_var

            session.add(User(name="Commit Test User"))
            await session.flush()

            assert stick_to_primary_var.get()

            await session.commit()

            assert not stick_to_primary_var.get()

        finally:
            await session.close()
            reset_routing_context()

        await async_session_maker.close_all()


class TestRoutingSyncSessionMaker:
    """Integration tests for sync session maker."""

    def test_creates_sessions_with_routing(self, routing_test_db_paths: dict[str, Path]) -> None:
        """Test that session maker creates properly configured routing sessions."""
        config = RoutingConfig(
            primary_connection_string=f"sqlite:///{routing_test_db_paths['primary']}",
            read_replicas=[f"sqlite:///{routing_test_db_paths['replica1']}"],
        )

        maker = RoutingSyncSessionMaker(routing_config=config)

        try:
            session = maker()
            assert isinstance(session, RoutingSyncSession)
            assert session._default_engine is maker.primary_engine
        finally:
            maker.close_all()

    def test_round_robin_selector(self, routing_test_db_paths: dict[str, Path]) -> None:
        """Test that round-robin strategy is applied."""
        config = RoutingConfig(
            primary_connection_string=f"sqlite:///{routing_test_db_paths['primary']}",
            read_replicas=[
                f"sqlite:///{routing_test_db_paths['replica1']}",
                f"sqlite:///{routing_test_db_paths['replica2']}",
            ],
            routing_strategy=RoutingStrategy.ROUND_ROBIN,
        )

        maker = RoutingSyncSessionMaker(routing_config=config)

        try:
            assert maker._selectors["read"] is not None
            selector = maker._selectors["read"]
            assert isinstance(selector, RoundRobinSelector)

            first = selector.next()
            second = selector.next()
            third = selector.next()

            assert first == third
            assert first != second

        finally:
            maker.close_all()

    def test_random_selector(self, routing_test_db_paths: dict[str, Path]) -> None:
        """Test that random strategy is applied."""
        config = RoutingConfig(
            primary_connection_string=f"sqlite:///{routing_test_db_paths['primary']}",
            read_replicas=[
                f"sqlite:///{routing_test_db_paths['replica1']}",
                f"sqlite:///{routing_test_db_paths['replica2']}",
            ],
            routing_strategy=RoutingStrategy.RANDOM,
        )

        maker = RoutingSyncSessionMaker(routing_config=config)

        try:
            assert isinstance(maker._selectors["read"], RandomSelector)
        finally:
            maker.close_all()


class TestRoutingAsyncSessionMaker:
    """Integration tests for async session maker."""

    @pytest.mark.asyncio
    async def test_creates_async_sessions_with_routing(self, routing_test_db_paths: dict[str, Path]) -> None:
        """Test that async session maker creates properly configured sessions."""
        config = RoutingConfig(
            primary_connection_string=f"sqlite+aiosqlite:///{routing_test_db_paths['primary']}",
            read_replicas=[f"sqlite+aiosqlite:///{routing_test_db_paths['replica1']}"],
        )

        maker = RoutingAsyncSessionMaker(routing_config=config)

        try:
            session = maker()
            assert isinstance(session, RoutingAsyncSession)
            assert session.primary_engine is maker.primary_engine
        finally:
            await maker.close_all()


class TestContextIsolation:
    """Tests for context variable isolation between concurrent requests."""

    def test_sync_context_isolation(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that context variables are isolated per execution context."""
        from concurrent.futures import ThreadPoolExecutor
        from threading import Barrier

        results: dict[str, bool] = {}
        barrier = Barrier(2)

        def task1() -> None:
            reset_routing_context()
            session = sync_session_maker()
            try:
                session.add(User(name="Task 1"))
                session.flush()

                from advanced_alchemy.routing.context import stick_to_primary_var

                barrier.wait()
                results["task1_sticky"] = stick_to_primary_var.get()
            finally:
                session.rollback()
                session.close()
                reset_routing_context()

        def task2() -> None:
            reset_routing_context()
            session = sync_session_maker()
            try:
                barrier.wait()
                from advanced_alchemy.routing.context import stick_to_primary_var

                results["task2_sticky"] = stick_to_primary_var.get()
            finally:
                session.close()
                reset_routing_context()

        with ThreadPoolExecutor(max_workers=2) as executor:
            f1 = executor.submit(task1)
            f2 = executor.submit(task2)
            f1.result()
            f2.result()

        assert results["task1_sticky"] is True
        assert results["task2_sticky"] is False

    @pytest.mark.asyncio
    async def test_async_context_isolation(self, async_session_maker: RoutingAsyncSessionMaker) -> None:
        """Test that async context variables are isolated per task."""
        async with async_session_maker.primary_engine.begin() as conn:
            await conn.run_sync(RoutingTestBase.metadata.create_all)

        results: dict[str, bool] = {}
        event = asyncio.Event()

        async def task1() -> None:
            reset_routing_context()
            session = async_session_maker()
            try:
                session.add(User(name="Async Task 1"))
                await session.flush()

                from advanced_alchemy.routing.context import stick_to_primary_var

                event.set()
                await asyncio.sleep(0.1)
                results["task1_sticky"] = stick_to_primary_var.get()
            finally:
                await session.rollback()
                await session.close()
                reset_routing_context()

        async def task2() -> None:
            reset_routing_context()
            session = async_session_maker()
            try:
                await event.wait()

                from advanced_alchemy.routing.context import stick_to_primary_var

                results["task2_sticky"] = stick_to_primary_var.get()
            finally:
                await session.close()
                reset_routing_context()

        await asyncio.gather(task1(), task2())

        assert results["task1_sticky"] is True
        assert results["task2_sticky"] is False

        await async_session_maker.close_all()


class TestEdgeCases:
    """Edge case tests for routing functionality."""

    def test_no_replicas_routes_to_primary(self, routing_test_db_paths: dict[str, Path]) -> None:
        """Test that without replicas, all operations go to primary."""
        config = RoutingConfig(
            primary_connection_string=f"sqlite:///{routing_test_db_paths['primary']}",
            read_replicas=[],
        )

        maker = RoutingSyncSessionMaker(routing_config=config)
        RoutingTestBase.metadata.create_all(maker.primary_engine)

        try:
            reset_routing_context()
            session = maker()

            with Session(maker.primary_engine) as direct_session:
                direct_session.add(User(id="no-replica-test", name="No Replica User"))
                direct_session.commit()

            result = session.execute(select(User).where(User.id == "no-replica-test")).scalar_one_or_none()
            assert result is not None
            assert result.name == "No Replica User"

        finally:
            session.close()
            maker.close_all()
            reset_routing_context()

    def test_routing_disabled(self, routing_test_db_paths: dict[str, Path]) -> None:
        """Test that disabled routing sends all operations to primary."""
        config = RoutingConfig(
            primary_connection_string=f"sqlite:///{routing_test_db_paths['primary']}",
            read_replicas=[f"sqlite:///{routing_test_db_paths['replica1']}"],
            enabled=False,
        )

        maker = RoutingSyncSessionMaker(routing_config=config)
        RoutingTestBase.metadata.create_all(maker.primary_engine)

        try:
            reset_routing_context()
            session = maker()

            with Session(maker.primary_engine) as direct_session:
                direct_session.add(User(id="disabled-test", name="Disabled Routing User"))
                direct_session.commit()

            result = session.execute(select(User).where(User.id == "disabled-test")).scalar_one_or_none()
            assert result is not None

        finally:
            session.close()
            maker.close_all()
            reset_routing_context()

    def test_for_update_routes_to_primary(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that FOR UPDATE queries route to primary."""
        reset_routing_context()

        with Session(sync_session_maker.primary_engine) as direct_session:
            direct_session.add(User(id="for-update-test", name="For Update User"))
            direct_session.commit()

        session = sync_session_maker()

        try:
            stmt = select(User).where(User.id == "for-update-test").with_for_update()
            result = session.execute(stmt).scalar_one_or_none()
            assert result is not None
            assert result.name == "For Update User"

        finally:
            session.close()
            reset_routing_context()


class TestNestedContexts:
    """Tests for nested context manager behavior."""

    def test_nested_primary_contexts(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that nested primary_context managers work correctly."""
        reset_routing_context()

        with primary_context():
            from advanced_alchemy.routing.context import force_primary_var

            assert force_primary_var.get()

            with primary_context():
                assert force_primary_var.get()

            assert force_primary_var.get()

        assert not force_primary_var.get()
        reset_routing_context()

    def test_nested_replica_contexts(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test that nested replica_context managers work correctly."""
        reset_routing_context()
        session = sync_session_maker()

        try:
            session.add(User(name="Sticky User"))
            session.flush()

            from advanced_alchemy.routing.context import (
                force_primary_var,
                stick_to_primary_var,
            )

            assert stick_to_primary_var.get()

            with replica_context():
                assert not force_primary_var.get()

                with replica_context():
                    assert not force_primary_var.get()

                assert not force_primary_var.get()

        finally:
            session.rollback()
            session.close()
            reset_routing_context()

    def test_mixed_contexts(self, sync_session_maker: RoutingSyncSessionMaker) -> None:
        """Test mixing primary_context and replica_context."""
        reset_routing_context()

        from advanced_alchemy.routing.context import force_primary_var

        with primary_context():
            assert force_primary_var.get()

            with replica_context():
                assert not force_primary_var.get()

            assert force_primary_var.get()

        assert not force_primary_var.get()
        reset_routing_context()



"""Integration tests for repository caching with bind_group support.

These tests verify that cache keys properly include bind_group for multi-master
database configurations, preventing data leaks between database shards.
"""

import asyncio
from collections.abc import Generator
from typing import TYPE_CHECKING, Any, Optional, cast

import pytest
from sqlalchemy import Engine, String, event
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from advanced_alchemy.base import UUIDBase
from advanced_alchemy.cache import setup_cache_listeners
from advanced_alchemy.cache.config import CacheConfig
from advanced_alchemy.cache.manager import DOGPILE_CACHE_INSTALLED, CacheManager
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository

if TYPE_CHECKING:
    pass

pytestmark = [
    pytest.mark.integration,
    pytest.mark.xdist_group("cache"),
]


@pytest.fixture(scope="module", autouse=True)
def _setup_cache_listeners() -> Generator[None, None, None]:
    """Set up global cache listeners for all tests in this module."""
    setup_cache_listeners()
    yield


# Module-level cache for model and counter for unique names
_model_cache: dict[str, type] = {}
_class_counter = 0


def get_bind_group_author_model(engine_dialect_name: str, worker_id: str) -> type[DeclarativeBase]:
    """Create appropriate model for bind_group tests."""
    global _class_counter
    cache_key = f"bind_group_author_{worker_id}_{engine_dialect_name}"

    if cache_key not in _model_cache:

        class TestBase(DeclarativeBase):
            pass

        _class_counter += 1
        unique_suffix = f"{_class_counter}_{worker_id}_{engine_dialect_name}"

        class_name = f"BindGroupAuthor_{unique_suffix}"

        BindGroupAuthor = type(
            class_name,
            (UUIDBase, TestBase),
            {
                "__tablename__": f"test_bind_group_authors_{worker_id}_{engine_dialect_name}",
                "__mapper_args__": {"concrete": True},
                "__module__": __name__,
                "name": mapped_column(String(length=100)),
                "bio": mapped_column(String(length=500), nullable=True),
                "__annotations__": {"name": Mapped[str], "bio": Mapped[Optional[str]]},
            },
        )

        _model_cache[cache_key] = BindGroupAuthor

    return _model_cache[cache_key]


def get_worker_id(request: pytest.FixtureRequest) -> str:
    """Get worker ID for pytest-xdist or 'master' for single process."""
    workerinput = getattr(request.config, "workerinput", None)
    if isinstance(workerinput, dict):
        return cast("str", workerinput.get("workerid", "master"))
    return "master"


@pytest.fixture
def memory_cache_manager() -> CacheManager:
    """Create a CacheManager with memory backend for testing."""
    config = CacheConfig(
        backend="dogpile.cache.memory",
        expiration_time=300,
        key_prefix="test_bind_group:",
    )
    return CacheManager(config)


# ============================================================================
# Cache Invalidation Tracker Tests
# ============================================================================


@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
def test_cache_invalidation_tracker_add_invalidation_with_bind_group(memory_cache_manager: CacheManager) -> None:
    """Test CacheInvalidationTracker stores bind_group with invalidations."""
    from advanced_alchemy._listeners import CacheInvalidationTracker

    tracker = CacheInvalidationTracker(memory_cache_manager)

    # Add invalidations with different bind_groups
    tracker.add_invalidation("model_a", "id1", bind_group=None)
    tracker.add_invalidation("model_a", "id2", bind_group="shard_a")
    tracker.add_invalidation("model_b", "id3", bind_group="shard_b")

    # Verify pending invalidations stored correctly
    assert len(tracker._pending_invalidations) == 3
    assert ("model_a", "id1", None) in tracker._pending_invalidations
    assert ("model_a", "id2", "shard_a") in tracker._pending_invalidations
    assert ("model_b", "id3", "shard_b") in tracker._pending_invalidations

    # Verify model bumps queued
    assert "model_a" in tracker._pending_model_bumps
    assert "model_b" in tracker._pending_model_bumps


@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
def test_cache_invalidation_tracker_rollback_clears_pending(memory_cache_manager: CacheManager) -> None:
    """Test CacheInvalidationTracker rollback clears all pending invalidations."""
    from advanced_alchemy._listeners import CacheInvalidationTracker

    tracker = CacheInvalidationTracker(memory_cache_manager)

    tracker.add_invalidation("model", "id1", bind_group="shard_a")
    tracker.add_invalidation("model", "id2", bind_group="shard_b")

    # Verify pending invalidations exist
    assert len(tracker._pending_invalidations) == 2
    assert len(tracker._pending_model_bumps) == 1

    # Rollback should clear everything
    tracker.rollback()

    assert len(tracker._pending_invalidations) == 0
    assert len(tracker._pending_model_bumps) == 0


# ============================================================================
# Async Repository Tests with bind_group
# ============================================================================


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_async_repository_get_with_bind_group_uses_separate_cache(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test async repository get() with bind_group uses separate cache entries."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_bind", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemyAsyncRepository[Any]):
                model_type = BindGroupAuthor

            repo = BindGroupAuthorRepository(session=session, cache_manager=memory_cache_manager, auto_expunge=True)

            # Create an author
            author = BindGroupAuthor(name="Test Author", bio="Bio")
            await repo.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Get without bind_group - should cache to default key
            author1 = await repo.get(author_id)
            assert author1.name == "Test Author"

            # Verify default cache was populated
            cached_default = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group=None
            )
            assert cached_default is not None

            # Get with bind_group="shard_a" - should cache to shard_a key
            author2 = await repo.get(author_id, bind_group="shard_a")
            assert author2.name == "Test Author"

            # Verify shard_a cache was populated
            cached_shard_a = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="shard_a"
            )
            assert cached_shard_a is not None

            # Invalidate only shard_a cache
            memory_cache_manager.invalidate_entity_sync(table_name, author_id, bind_group="shard_a")

            # Verify default cache still exists
            cached_default_after = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group=None
            )
            assert cached_default_after is not None, "Default cache should still exist"

            # Verify shard_a cache was invalidated
            cached_shard_a_after = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="shard_a"
            )
            assert cached_shard_a_after is None, "shard_a cache should be invalidated"

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_async_repository_singleflight_key_includes_bind_group(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test that singleflight deduplication includes bind_group in key."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_sf", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    query_count = 0

    def before_cursor_execute(_conn: object, _cursor: object, statement: str, *_: object) -> None:
        nonlocal query_count
        if statement.lstrip().upper().startswith("SELECT"):
            query_count += 1

    event.listen(aiosqlite_engine.sync_engine, "before_cursor_execute", before_cursor_execute)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemyAsyncRepository[Any]):
                model_type = BindGroupAuthor

            repo = BindGroupAuthorRepository(session=session, cache_manager=memory_cache_manager, auto_expunge=True)

            author = BindGroupAuthor(name="SF Author")
            await repo.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Invalidate cache to force misses
            memory_cache_manager.invalidate_entity_sync(table_name, author_id, bind_group=None)
            memory_cache_manager.invalidate_entity_sync(table_name, author_id, bind_group="shard_a")

            # Concurrent gets without bind_group should coalesce
            query_count = 0
            results_default = await asyncio.gather(*[repo.get(author_id) for _ in range(5)])
            assert all(r.id == author_id for r in results_default)
            queries_default = query_count

            # Concurrent gets with bind_group="shard_a" should coalesce separately
            query_count = 0
            results_shard_a = await asyncio.gather(*[repo.get(author_id, bind_group="shard_a") for _ in range(5)])
            assert all(r.id == author_id for r in results_shard_a)
            queries_shard_a = query_count

            # Both should have coalesced to 1 query each
            assert queries_default == 1, "Default queries should coalesce"
            assert queries_shard_a == 1, "shard_a queries should coalesce"

    finally:
        event.remove(aiosqlite_engine.sync_engine, "before_cursor_execute", before_cursor_execute)
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_async_repository_get_bypasses_cache_when_disabled(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test async repository get() with use_cache=False bypasses cache regardless of bind_group."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_bypass", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemyAsyncRepository[Any]):
                model_type = BindGroupAuthor

            repo = BindGroupAuthorRepository(session=session, cache_manager=memory_cache_manager, auto_expunge=True)

            author = BindGroupAuthor(name="No Cache Author")
            await repo.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Get with cache disabled and bind_group
            await repo.get(author_id, use_cache=False, bind_group="shard_a")

            # Cache should not be populated
            cached = memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_a")
            assert cached is None, "Cache should not be populated when use_cache=False"

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


# ============================================================================
# Sync Repository Tests with bind_group
# ============================================================================


@pytest.mark.sqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
def test_sync_repository_get_with_bind_group_uses_separate_cache(
    sqlite_engine: Engine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test sync repository get() with bind_group uses separate cache entries."""
    from sqlalchemy.orm import sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("sqlite_bind", worker_id)

    BindGroupAuthor.metadata.create_all(sqlite_engine)

    try:
        session_factory = sessionmaker(sqlite_engine)
        with session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemySyncRepository[Any]):
                model_type = BindGroupAuthor

            repo = BindGroupAuthorRepository(session=session, cache_manager=memory_cache_manager, auto_expunge=True)

            # Create an author
            author = BindGroupAuthor(name="Test Author", bio="Bio")
            repo.add(author)
            session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Get without bind_group
            author1 = repo.get(author_id)
            assert author1.name == "Test Author"

            # Verify default cache was populated
            cached_default = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group=None
            )
            assert cached_default is not None

            # Get with bind_group="shard_a"
            author2 = repo.get(author_id, bind_group="shard_a")
            assert author2.name == "Test Author"

            # Verify shard_a cache was populated
            cached_shard_a = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="shard_a"
            )
            assert cached_shard_a is not None

            # Invalidate only shard_a
            memory_cache_manager.invalidate_entity_sync(table_name, author_id, bind_group="shard_a")

            # Verify default still cached, shard_a invalidated
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group=None)
                is not None
            )
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_a")
                is None
            )

    finally:
        BindGroupAuthor.metadata.drop_all(sqlite_engine)


@pytest.mark.sqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
def test_sync_repository_get_bypasses_cache_when_disabled(
    sqlite_engine: Engine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test sync repository get() with use_cache=False bypasses cache regardless of bind_group."""
    from sqlalchemy.orm import sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("sqlite_bypass", worker_id)

    BindGroupAuthor.metadata.create_all(sqlite_engine)

    try:
        session_factory = sessionmaker(sqlite_engine)
        with session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemySyncRepository[Any]):
                model_type = BindGroupAuthor

            repo = BindGroupAuthorRepository(session=session, cache_manager=memory_cache_manager, auto_expunge=True)

            author = BindGroupAuthor(name="No Cache")
            repo.add(author)
            session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Get with cache disabled and bind_group
            repo.get(author_id, use_cache=False, bind_group="shard_a")

            # Cache should not be populated
            cached = memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_a")
            assert cached is None, "Cache should not be populated when use_cache=False"

    finally:
        BindGroupAuthor.metadata.drop_all(sqlite_engine)


# ============================================================================
# Repository default bind_group Tests
# ============================================================================


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_async_repository_uses_default_bind_group_for_cache(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test that repository uses default bind_group from constructor for cache keys."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_default", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemyAsyncRepository[Any]):
                model_type = BindGroupAuthor

            # Create repo with default bind_group via constructor parameter
            repo = BindGroupAuthorRepository(
                session=session,
                cache_manager=memory_cache_manager,
                auto_expunge=True,
                bind_group="default_shard",
            )

            author = BindGroupAuthor(name="Default Shard Author")
            await repo.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Get without explicit bind_group - should use default from constructor
            await repo.get(author_id)

            # Verify cache was populated for default_shard bind_group
            cached = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="default_shard"
            )
            assert cached is not None, "Cache should use default bind_group from constructor"

            # Verify no cache for None bind_group
            cached_none = memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group=None)
            assert cached_none is None, "No cache should exist for None bind_group"

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_async_repository_explicit_bind_group_overrides_default(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test that explicit bind_group parameter overrides repository default."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_override", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:

            class BindGroupAuthorRepository(SQLAlchemyAsyncRepository[Any]):
                model_type = BindGroupAuthor

            # Create repo with default bind_group via constructor
            repo = BindGroupAuthorRepository(
                session=session,
                cache_manager=memory_cache_manager,
                auto_expunge=True,
                bind_group="default_shard",
            )

            author = BindGroupAuthor(name="Override Test")
            await repo.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Get with explicit bind_group override
            await repo.get(author_id, bind_group="override_shard")

            # Verify cache was populated for override_shard, not default_shard
            cached_override = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="override_shard"
            )
            cached_default = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="default_shard"
            )

            assert cached_override is not None, "Cache should be for override_shard"
            assert cached_default is None, "No cache for default_shard when overridden"

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


# ============================================================================
# Cache Manager Direct Tests (using model instances)
# ============================================================================


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_cache_manager_entity_methods_with_bind_group(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test CacheManager entity methods correctly handle bind_group in cache keys."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_cm", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:
            # Create a model instance
            author = BindGroupAuthor(name="Cache Manager Test")
            session.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Test set_entity_sync with different bind_groups
            memory_cache_manager.set_entity_sync(table_name, author_id, author, bind_group=None)

            # Modify for shard_a (simulate different data in shard)
            author.name = "Shard A Data"
            memory_cache_manager.set_entity_sync(table_name, author_id, author, bind_group="shard_a")

            # Modify for shard_b
            author.name = "Shard B Data"
            memory_cache_manager.set_entity_sync(table_name, author_id, author, bind_group="shard_b")

            # Test get_entity_sync returns correct cached entity per bind_group
            cached_default = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group=None
            )
            cached_shard_a = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="shard_a"
            )
            cached_shard_b = memory_cache_manager.get_entity_sync(
                table_name, author_id, BindGroupAuthor, bind_group="shard_b"
            )

            assert cached_default is not None
            assert cached_shard_a is not None
            assert cached_shard_b is not None

            # Names should reflect what was cached at the time
            assert cached_default.name == "Cache Manager Test"
            assert cached_shard_a.name == "Shard A Data"
            assert cached_shard_b.name == "Shard B Data"

            # Test invalidate_entity_sync only affects specific bind_group
            memory_cache_manager.invalidate_entity_sync(table_name, author_id, bind_group="shard_a")

            # Verify only shard_a was invalidated
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group=None)
                is not None
            )
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_a")
                is None
            )
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_b")
                is not None
            )

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_cache_manager_async_entity_methods_with_bind_group(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test CacheManager async entity methods correctly handle bind_group in cache keys."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_cm_async", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:
            author = BindGroupAuthor(name="Async Cache Test")
            session.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Test async set/get with bind_groups
            await memory_cache_manager.set_entity_async(table_name, author_id, author, bind_group=None)

            author.name = "Shard A Async"
            await memory_cache_manager.set_entity_async(table_name, author_id, author, bind_group="shard_a")

            # Verify both cached correctly
            cached_default = await memory_cache_manager.get_entity_async(
                table_name, author_id, BindGroupAuthor, bind_group=None
            )
            cached_shard_a = await memory_cache_manager.get_entity_async(
                table_name, author_id, BindGroupAuthor, bind_group="shard_a"
            )

            assert cached_default is not None
            assert cached_shard_a is not None
            assert cached_default.name == "Async Cache Test"
            assert cached_shard_a.name == "Shard A Async"

            # Test async invalidation
            await memory_cache_manager.invalidate_entity_async(table_name, author_id, bind_group="shard_a")

            # Verify only shard_a was invalidated
            assert (
                await memory_cache_manager.get_entity_async(table_name, author_id, BindGroupAuthor, bind_group=None)
                is not None
            )
            assert (
                await memory_cache_manager.get_entity_async(
                    table_name, author_id, BindGroupAuthor, bind_group="shard_a"
                )
                is None
            )

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


# ============================================================================
# Cache Invalidation Tracker with Commit Tests
# ============================================================================


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_cache_invalidation_tracker_commit_with_bind_group(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test CacheInvalidationTracker commit properly invalidates with bind_group."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    from advanced_alchemy._listeners import CacheInvalidationTracker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_tracker", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:
            author = BindGroupAuthor(name="Tracker Test")
            session.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Pre-populate cache for different bind_groups
            memory_cache_manager.set_entity_sync(table_name, author_id, author, bind_group=None)
            memory_cache_manager.set_entity_sync(table_name, author_id, author, bind_group="shard_a")
            memory_cache_manager.set_entity_sync(table_name, author_id, author, bind_group="shard_b")

            # Create tracker and add invalidation for shard_a only
            tracker = CacheInvalidationTracker(memory_cache_manager)
            tracker.add_invalidation(table_name, author_id, bind_group="shard_a")

            # Commit the tracker (sync)
            tracker.commit()

            # Verify only shard_a was invalidated
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group=None)
                is not None
            )
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_a")
                is None
            )
            assert (
                memory_cache_manager.get_entity_sync(table_name, author_id, BindGroupAuthor, bind_group="shard_b")
                is not None
            )

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)


@pytest.mark.asyncio
@pytest.mark.aiosqlite
@pytest.mark.skipif(not DOGPILE_CACHE_INSTALLED, reason="dogpile.cache not installed")
async def test_cache_invalidation_tracker_async_commit_with_bind_group(
    aiosqlite_engine: AsyncEngine,
    memory_cache_manager: CacheManager,
    request: pytest.FixtureRequest,
) -> None:
    """Test CacheInvalidationTracker async commit properly invalidates with bind_group."""
    from sqlalchemy.ext.asyncio import AsyncSession as AS
    from sqlalchemy.ext.asyncio import async_sessionmaker

    from advanced_alchemy._listeners import CacheInvalidationTracker

    worker_id = get_worker_id(request)
    BindGroupAuthor = get_bind_group_author_model("aiosqlite_tracker_async", worker_id)

    async with aiosqlite_engine.begin() as conn:
        await conn.run_sync(BindGroupAuthor.metadata.create_all)

    try:
        async_session_factory = async_sessionmaker(aiosqlite_engine, class_=AS, expire_on_commit=False)
        async with async_session_factory() as session:
            author = BindGroupAuthor(name="Async Tracker Test")
            session.add(author)
            await session.commit()

            author_id = author.id
            table_name = BindGroupAuthor.__tablename__

            # Pre-populate cache
            await memory_cache_manager.set_entity_async(table_name, author_id, author, bind_group=None)
            await memory_cache_manager.set_entity_async(table_name, author_id, author, bind_group="shard_a")

            # Create tracker and add invalidation
            tracker = CacheInvalidationTracker(memory_cache_manager)
            tracker.add_invalidation(table_name, author_id, bind_group="shard_a")

            # Async commit
            await tracker.commit_async()

            # Verify only shard_a was invalidated
            assert (
                await memory_cache_manager.get_entity_async(table_name, author_id, BindGroupAuthor, bind_group=None)
                is not None
            )
            assert (
                await memory_cache_manager.get_entity_async(
                    table_name, author_id, BindGroupAuthor, bind_group="shard_a"
                )
                is None
            )

    finally:
        async with aiosqlite_engine.begin() as conn:
            await conn.run_sync(BindGroupAuthor.metadata.drop_all)



"""Sphinx extension for changelog and change directives."""

# ruff: noqa: PLR0911, ARG001
from __future__ import annotations

import ast
import importlib
import inspect
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator

    from docutils.nodes import Element, Node
    from sphinx.addnodes import pending_xref
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment


def _get_module_ast(source_file: str) -> ast.AST | ast.Module:
    return ast.parse(Path(source_file).read_text(encoding="utf-8"))


def _get_import_nodes(nodes: list[ast.stmt]) -> Generator[ast.Import | ast.ImportFrom, None, None]:
    for node in nodes:
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            yield node
        elif isinstance(node, ast.If) and getattr(node.test, "id", None) == "TYPE_CHECKING":
            yield from _get_import_nodes(node.body)


def get_module_global_imports(module_import_path: str, reference_target_source_obj: str) -> set[str]:
    """Return a set of names that are imported globally within the containing module of ``reference_target_source_obj``,
    including imports in ``if TYPE_CHECKING`` blocks.
    """
    module = importlib.import_module(module_import_path)
    obj = getattr(module, reference_target_source_obj)
    tree = _get_module_ast(inspect.getsourcefile(obj))  # type: ignore[arg-type]  # pyright: ignore[reportArgumentType]

    import_nodes = _get_import_nodes(tree.body)  # type: ignore[attr-defined]
    return {path.asname or path.name for import_node in import_nodes for path in import_node.names}


def _resolve_local_reference(module_path: str, target: str) -> bool:
    """Attempt to resolve a reference within the local codebase.

    Args:
        module_path: The module path (e.g., 'advanced_alchemy.base')
        target: The target class/attribute name

    Returns:
        bool: True if reference exists, False otherwise
    """
    try:
        module = importlib.import_module(module_path)
        if "." in target:
            # Handle fully qualified names (e.g., advanced_alchemy.base.BasicAttributes)
            parts = target.split(".")
            current = module
            for part in parts:
                current = getattr(current, part)
            return True
        return hasattr(module, target)
    except (ImportError, AttributeError):
        return False


def _resolve_sqlalchemy_reference(target: str) -> bool:
    """Attempt to resolve SQLAlchemy references.

    Args:
        target: The target class/attribute name

    Returns:
        bool: True if reference exists, False otherwise
    """
    try:
        import sqlalchemy

        if "." in target:
            # Handle nested attributes (e.g., Connection.in_transaction)
            obj_name, attr_name = target.rsplit(".", 1)
            obj = getattr(sqlalchemy, obj_name)
            return hasattr(obj, attr_name)
        return hasattr(sqlalchemy, target)
    except (ImportError, AttributeError):
        return False


def _resolve_litestar_reference(target: str) -> bool:
    """Attempt to resolve Litestar references.

    Args:
        target: The target class/attribute name

    Returns:
        bool: True if reference exists, False otherwise
    """
    try:
        import litestar
        from litestar import datastructures

        # Handle common Litestar classes
        if target in {"Litestar", "State", "Scope", "Message", "AppConfig", "BeforeMessageSendHookHandler"}:
            return True
        if target.startswith("datastructures."):
            _, attr = target.split(".")
            return hasattr(datastructures, attr)
        if target.startswith("config.app."):
            return True  # These are valid Litestar config references
        return hasattr(litestar, target)
    except ImportError:
        return False


def _resolve_sqlalchemy_type_reference(target: str) -> bool:
    """Attempt to resolve SQLAlchemy type references.

    Args:
        target: The target class/attribute name

    Returns:
        bool: True if reference exists, False otherwise
    """
    try:
        from sqlalchemy import types as sa_types

        type_classes = {
            "TypeEngine",
            "TypeDecorator",
            "UserDefinedType",
            "ExternalType",
            "Dialect",
            "_types.TypeDecorator",
        }

        if target in type_classes:
            return True
        if target.startswith("_types."):
            _, attr = target.split(".")
            return hasattr(sa_types, attr)
        return hasattr(sa_types, target)
    except ImportError:
        return False


def _resolve_advanced_alchemy_reference(target: str, module: str) -> bool:
    """Attempt to resolve Advanced Alchemy references.

    Args:
        target: The target class/attribute name
        module: The current module context

    Returns:
        bool: True if reference exists, False otherwise
    """
    # Handle base module references
    base_classes = {
        "BasicAttributes",
        "CommonTableAttributes",
        "AuditColumns",
        "BigIntPrimaryKey",
        "UUIDPrimaryKey",
        "UUIDv6PrimaryKey",
        "UUIDv7PrimaryKey",
        "NanoIDPrimaryKey",
        "Empty",
        "TableArgsType",
        "DeclarativeBase",
    }

    # Handle config module references
    config_classes = {
        "EngineT",
        "SessionT",
        "SessionMakerT",
        "ConnectionT",
        "GenericSessionConfig",
        "GenericAlembicConfig",
    }

    func_references = {"repository.SQLAlchemyAsyncRepositoryProtocol.add_many"}

    # Handle type module references
    type_classes = {"DateTimeUTC", "ORA_JSONB", "GUID", "EncryptedString", "EncryptedText"}

    if target in base_classes or target in config_classes or target in type_classes:
        return True

    # Handle fully qualified references
    if target.startswith("advanced_alchemy."):
        parts = target.split(".")
        if parts[-1] in base_classes | config_classes | type_classes | func_references:
            return True

    # Handle module-relative references
    return bool(module.startswith("advanced_alchemy."))


def _resolve_serialization_reference(target: str) -> bool:
    """Attempt to resolve serialization-related references.

    Args:
        target: The target class/attribute name

    Returns:
        bool: True if reference exists, False otherwise
    """
    serialization_attrs = {"decode_json", "encode_json", "serialization.decode_json", "serialization.encode_json"}
    return target in serialization_attrs


def _resolve_click_reference(target: str) -> bool:
    """Attempt to resolve Click references.

    Args:
        target: The target class/attribute name

    Returns:
        bool: True if reference exists, False otherwise
    """
    try:
        import click

        if target == "Group":
            return True
        return hasattr(click, target)
    except ImportError:
        return False


def on_warn_missing_reference(app: Sphinx, domain: str, node: Node) -> bool | None:
    if node.tagname != "pending_xref":  # type: ignore[attr-defined]
        return None

    if not hasattr(node, "attributes"):
        return None

    # Wrap the main logic in a try-except to catch potential AttributeErrors (e.g., startswith on None)
    try:
        attributes = node.attributes  # type: ignore[attr-defined,unused-ignore]
        target = attributes["reftarget"]  # pyright: ignore
        ref_type = attributes.get("reftype")  # pyright: ignore
        module = attributes.get("py:module", "")  # pyright: ignore

        # Handle TypeVar references
        if hasattr(target, "__class__") and target.__class__.__name__ == "TypeVar":  # pyright: ignore
            return True

        # Handle Advanced Alchemy references
        if _resolve_advanced_alchemy_reference(target, module):  # pyright: ignore
            return True

        # Handle SQLAlchemy type system references
        if ref_type in {"class", "meth", "attr"} and any(
            x in target for x in ["TypeDecorator", "TypeEngine", "Dialect", "ExternalType", "UserDefinedType"]
        ):
            return _resolve_sqlalchemy_type_reference(target)  # pyright: ignore

        # Handle SQLAlchemy core references
        if (isinstance(target, str) and target.startswith("sqlalchemy.")) or (
            ref_type in {"class", "attr", "obj", "meth"}
            and target
            in {
                "Engine",
                "Session",
                "Connection",
                "MetaData",
                "AsyncSession",
                "AsyncEngine",
                "AsyncConnection",
                "sessionmaker",
                "async_sessionmaker",
            }
        ):
            # Ensure target is string before replace
            clean_target = target.replace("sqlalchemy.", "") if isinstance(target, str) else ""
            if clean_target and _resolve_sqlalchemy_reference(clean_target):
                return True

        # Handle Litestar references
        if ref_type in {"class", "obj"} and (
            (isinstance(target, str) and target.startswith(("datastructures.", "config.app.")))
            or target
            in {
                "Litestar",
                "State",
                "Scope",
                "Message",
                "AppConfig",
                "BeforeMessageSendHookHandler",
                "FieldDefinition",
                "ImproperConfigurationError",
            }
        ):
            return _resolve_litestar_reference(target)  # pyright: ignore

        # Handle serialization references
        if ref_type in {"attr", "meth"} and _resolve_serialization_reference(target):  # pyright: ignore
            return True

        # Handle Click references
        if ref_type == "class" and _resolve_click_reference(target):  # pyright: ignore
            return True

    except AttributeError:
        # Catch the specific error (likely startswith on None) and allow Sphinx to handle the warning normally
        return None

    return None


def on_missing_reference(app: Sphinx, env: BuildEnvironment, node: pending_xref, contnode: Element) -> Element | None:
    """Handle missing references by attempting to resolve them through different methods.

    Args:
        app: The Sphinx application instance
        env: The Sphinx build environment
        node: The pending cross-reference node
        contnode: The content node

    Returns:
        Element | None: The resolved reference node if found, None otherwise
    """
    if not hasattr(node, "attributes"):
        return None

    attributes = node.attributes  # type: ignore[attr-defined,unused-ignore]
    target = attributes["reftarget"]

    # Remove this check since it's causing issues
    if not isinstance(target, str):
        return None

    py_domain = env.domains["py"]

    # autodoc sometimes incorrectly resolves these types, so we try to resolve them as py:data first and fall back to any
    new_node = py_domain.resolve_xref(env, node["refdoc"], app.builder, "data", target, node, contnode)
    if new_node is None:
        resolved_xrefs = py_domain.resolve_any_xref(env, node["refdoc"], app.builder, target, node, contnode)
        for ref in resolved_xrefs:
            if ref:
                return ref[1]
    return new_node


def on_env_before_read_docs(app: Sphinx, env: BuildEnvironment, docnames: set[str]) -> None:
    tmp_examples_path = Path.cwd() / "docs/_build/_tmp_examples"
    tmp_examples_path.mkdir(exist_ok=True, parents=True)
    env.tmp_examples_path = tmp_examples_path  # type: ignore[attr-defined] # pyright: ignore[reportAttributeAccessIssue]


def setup(app: Sphinx) -> dict[str, bool]:
    app.connect("env-before-read-docs", on_env_before_read_docs)
    app.connect("missing-reference", on_missing_reference)
    app.connect("warn-missing-reference", on_warn_missing_reference)
    app.add_config_value("ignore_missing_refs", default={}, rebuild="")
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Tests for advanced_alchemy.operations module."""

from typing import Any

import pytest
from sqlalchemy import Column, Integer, MetaData, String, Table

from advanced_alchemy.operations import MergeStatement, OnConflictUpsert, validate_identifier


@pytest.fixture
def sample_table() -> Table:
    """Create a sample table for testing."""
    metadata = MetaData()
    return Table(
        "test_table",
        metadata,
        Column("id", Integer, primary_key=True),
        Column("key", String(50), nullable=False),
        Column("namespace", String(50), nullable=False),
        Column("value", String(255)),
    )


class TestOnConflictUpsert:
    """Test OnConflictUpsert operations."""

    def test_supports_native_upsert(self) -> None:
        """Test dialect support detection."""
        # Supported dialects
        assert OnConflictUpsert.supports_native_upsert("postgresql") is True
        assert OnConflictUpsert.supports_native_upsert("cockroachdb") is True
        assert OnConflictUpsert.supports_native_upsert("sqlite") is True
        assert OnConflictUpsert.supports_native_upsert("mysql") is True
        assert OnConflictUpsert.supports_native_upsert("mariadb") is True
        assert OnConflictUpsert.supports_native_upsert("duckdb") is True

        # Unsupported dialects
        assert OnConflictUpsert.supports_native_upsert("oracle") is False
        assert OnConflictUpsert.supports_native_upsert("mssql") is False
        assert OnConflictUpsert.supports_native_upsert("unknown") is False

    def test_create_postgresql_upsert(self, sample_table: Table) -> None:
        """Test PostgreSQL ON CONFLICT upsert generation."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}
        conflict_columns = ["key", "namespace"]

        upsert_stmt = OnConflictUpsert.create_upsert(
            table=sample_table,
            values=values,
            conflict_columns=conflict_columns,
            dialect_name="postgresql",
        )

        # Should return a PostgreSQL insert statement with ON CONFLICT
        # This is primarily testing that the method doesn't raise an exception
        # and returns the expected type
        assert upsert_stmt is not None
        assert hasattr(upsert_stmt, "on_conflict_do_update")

    def test_create_mysql_upsert(self, sample_table: Table) -> None:
        """Test MySQL ON DUPLICATE KEY UPDATE upsert generation."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}
        conflict_columns = ["key", "namespace"]

        upsert_stmt = OnConflictUpsert.create_upsert(
            table=sample_table,
            values=values,
            conflict_columns=conflict_columns,
            dialect_name="mysql",
        )

        # Should return a MySQL insert statement with ON DUPLICATE KEY UPDATE
        assert upsert_stmt is not None
        assert hasattr(upsert_stmt, "on_duplicate_key_update")

    def test_create_duckdb_upsert(self, sample_table: Table) -> None:
        """Test DuckDB ON CONFLICT upsert generation."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}
        conflict_columns = ["key", "namespace"]

        upsert_stmt = OnConflictUpsert.create_upsert(
            table=sample_table,
            values=values,
            conflict_columns=conflict_columns,
            dialect_name="duckdb",
        )

        # Should return a PostgreSQL-style insert statement with ON CONFLICT (DuckDB uses PostgreSQL syntax)
        assert upsert_stmt is not None
        assert hasattr(upsert_stmt, "on_conflict_do_update")

    def test_create_upsert_unsupported_dialect(self, sample_table: Table) -> None:
        """Test that unsupported dialects raise NotImplementedError."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}
        conflict_columns = ["key", "namespace"]

        with pytest.raises(NotImplementedError, match="Native upsert not supported for dialect 'oracle'"):
            OnConflictUpsert.create_upsert(
                table=sample_table,
                values=values,
                conflict_columns=conflict_columns,
                dialect_name="oracle",
            )

    def test_create_merge_upsert(self, sample_table: Table) -> None:
        """Test MERGE-based upsert generation."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}
        conflict_columns = ["key", "namespace"]

        # Test default (non-Oracle) MERGE
        merge_stmt, additional_params = OnConflictUpsert.create_merge_upsert(
            table=sample_table,
            values=values,
            conflict_columns=conflict_columns,
        )

        assert isinstance(merge_stmt, MergeStatement)
        assert merge_stmt.table == sample_table
        # Default dialect (None) uses PostgreSQL format with %(key)s notation
        assert "%(key)s" in str(merge_stmt.source) or ":key" in str(
            merge_stmt.source
        )  # Check for parameter placeholder
        assert "FROM DUAL" not in str(merge_stmt.source)  # Should not have FROM DUAL by default
        assert additional_params == {}  # No additional params for non-Oracle

        # Test Oracle-specific MERGE
        oracle_merge_stmt, oracle_additional_params = OnConflictUpsert.create_merge_upsert(
            table=sample_table,
            values=values,
            conflict_columns=conflict_columns,
            dialect_name="oracle",
        )

        assert isinstance(oracle_merge_stmt, MergeStatement)
        assert oracle_merge_stmt.table == sample_table
        assert ":key" in str(oracle_merge_stmt.source)  # Check for parameter placeholder
        assert "FROM DUAL" in str(oracle_merge_stmt.source)  # Oracle should have FROM DUAL
        # Additional params should be empty for tables without UUID primary keys
        assert isinstance(oracle_additional_params, dict)
        assert "SELECT" in str(merge_stmt.source)


class TestMergeStatement:
    """Test MergeStatement compilation."""

    def test_merge_statement_creation(self, sample_table: Table) -> None:
        """Test basic MergeStatement creation."""
        from sqlalchemy import bindparam, text

        source = "SELECT 'key1' as key, 'ns1' as namespace, 'value1' as value"
        on_condition = text("tgt.key = src.key AND tgt.namespace = src.namespace")
        when_matched_update: dict[str, Any] = {"value": bindparam("value")}
        when_not_matched_insert: dict[str, Any] = {
            "key": bindparam("key"),
            "namespace": bindparam("namespace"),
            "value": bindparam("value"),
        }

        merge_stmt = MergeStatement(
            table=sample_table,
            source=source,
            on_condition=on_condition,
            when_matched_update=when_matched_update,
            when_not_matched_insert=when_not_matched_insert,
        )

        assert merge_stmt.table == sample_table
        assert merge_stmt.source == source
        assert merge_stmt.on_condition == on_condition
        assert merge_stmt.when_matched_update == when_matched_update
        assert merge_stmt.when_not_matched_insert == when_not_matched_insert

    def test_compile_merge_default_raises_error(self, sample_table: Table) -> None:
        """Test that default compiler raises NotImplementedError."""
        from sqlalchemy import text

        from advanced_alchemy.operations import compile_merge_default

        merge_stmt = MergeStatement(
            table=sample_table,
            source="SELECT 1",
            on_condition=text("1=1"),
        )

        # Create a mock compiler for an unsupported dialect
        class MockDialect:
            name = "unsupported"

        class MockCompiler:
            dialect = MockDialect()

        compiler = MockCompiler()

        with pytest.raises(NotImplementedError, match="MERGE statement not supported for dialect 'unsupported'"):
            compile_merge_default(merge_stmt, compiler)  # type: ignore[arg-type]  # pyright: ignore


class TestIdentifierValidation:
    """Test identifier validation security feature."""

    def test_valid_identifiers(self) -> None:
        """Test that valid identifiers pass validation."""
        assert validate_identifier("user_id") == "user_id"
        assert validate_identifier("users_table", "table") == "users_table"
        assert validate_identifier("created_at", "column") == "created_at"
        assert validate_identifier("_private_field") == "_private_field"
        assert validate_identifier("table123") == "table123"

    def test_empty_identifier(self) -> None:
        """Test that empty identifiers are rejected."""
        with pytest.raises(ValueError, match="Empty identifier name"):
            validate_identifier("")

    def test_invalid_characters(self) -> None:
        """Test that identifiers with invalid characters are rejected."""
        invalid_names = [
            "user-id",  # hyphen
            "user.id",  # dot
            "user id",  # space
            "123user",  # starts with number
            "user;",  # semicolon
            "user'",  # quote
            "user`",  # backtick
            "drop table users; --",  # SQL injection attempt
        ]

        for name in invalid_names:
            with pytest.raises(ValueError, match=r"Invalid.*Only alphanumeric"):
                validate_identifier(name)

    def test_sql_keywords_allowed(self) -> None:
        """Test that SQL keywords are allowed as identifiers."""
        # SQL keywords should be allowed since they can be quoted in SQL
        keywords = ["select", "SELECT", "insert", "UPDATE", "delete", "DROP", "create", "ALTER", "truncate"]

        for keyword in keywords:
            # Should not raise an error
            assert validate_identifier(keyword) == keyword
            assert validate_identifier(keyword.lower()) == keyword.lower()
            assert validate_identifier(keyword.upper()) == keyword.upper()

    def test_identifier_type_in_error(self) -> None:
        """Test that identifier type appears in error messages."""
        with pytest.raises(ValueError, match="Empty column name"):
            validate_identifier("", "column")

        with pytest.raises(ValueError, match="Invalid table name"):
            validate_identifier("123invalid", "table")

    def test_upsert_with_validation(self, sample_table: Table) -> None:
        """Test that create_upsert validates identifiers when requested."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}

        # Should work with validation enabled for valid identifiers
        upsert_stmt = OnConflictUpsert.create_upsert(
            table=sample_table,
            values=values,
            conflict_columns=["key", "namespace"],
            update_columns=["value"],
            dialect_name="postgresql",
            validate_identifiers=True,
        )
        assert upsert_stmt is not None

    def test_merge_with_validation(self, sample_table: Table) -> None:
        """Test that create_merge_upsert validates identifiers when requested."""
        values = {"key": "test_key", "namespace": "test_ns", "value": "test_value"}

        # Should work with validation enabled for valid identifiers
        merge_stmt, _ = OnConflictUpsert.create_merge_upsert(
            table=sample_table,
            values=values,
            conflict_columns=["key", "namespace"],
            update_columns=["value"],
            dialect_name="oracle",
            validate_identifiers=True,
        )
        assert merge_stmt is not None


class TestStoreIntegration:
    """Test that the store can use the new operations."""

    def test_store_imports_operations(self) -> None:
        """Test that store successfully imports new operations."""
        from advanced_alchemy.extensions.litestar.store import SQLAlchemyStore
        from advanced_alchemy.operations import MergeStatement, OnConflictUpsert

        # This test passes if no import errors occur
        assert OnConflictUpsert is not None
        assert MergeStatement is not None
        assert SQLAlchemyStore is not None
